##########################################################################
# Copyright (c) 2010-2023 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
Markers
*******

:module: markers

:synopsis: define new pykiso-related markers.

"""

from _pytest.config import Config

from .utils import *


@export
def pytest_configure(config: Config):
    config.addinivalue_line(
        "markers",
        "test_ids(DUT1=['ABC123', 'DEF456'], DUT2=['123ABC']): marker for test IDs or test-related requirements that will appear in the report",
    )
    config.addinivalue_line(
        "markers",
        "tags(variant='DUT1', branch=['main', 'master']): marker to select specific test cases based on the tag name and value",
    )



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
pykiso Control Message Protocol
*******************************

:module: message

:synopsis: Message that will be send though the different agents

.. currentmodule:: message

"""

from __future__ import annotations

import enum
import itertools
import logging
import struct
from typing import Dict, Optional, Union

msg_cnt = itertools.cycle(range(256))  # Will be used as token. It increases each time a Message is created

log = logging.getLogger(__name__)


@enum.unique
class MessageType(enum.IntEnum):
    """List of messages allowed."""

    COMMAND = 0
    REPORT = 1
    ACK = 2
    LOG = 3


@enum.unique
class MessageCommandType(enum.IntEnum):
    """List of commands allowed."""

    # Ping
    PING = 0
    # Setups
    TEST_SECTION_SETUP = 1
    TEST_SUITE_SETUP = 2
    TEST_CASE_SETUP = 3
    # Runs
    TEST_SECTION_RUN = 11
    TEST_SUITE_RUN = 12
    TEST_CASE_RUN = 13
    # Teardowns
    TEST_SECTION_TEARDOWN = 21
    TEST_SUITE_TEARDOWN = 22
    TEST_CASE_TEARDOWN = 23
    # Abort
    ABORT = 99


@enum.unique
class MessageReportType(enum.IntEnum):
    """List of possible received messages."""

    TEST_PASS = 0
    TEST_FAILED = 1
    TEST_NOT_IMPLEMENTED = 2

    def __str__(self) -> str:
        return f"{self.__class__.__name__}.{self.name}"


@enum.unique
class MessageAckType(enum.IntEnum):
    """List of possible received messages."""

    ACK = 0
    NACK = 1


@enum.unique
class TlvKnownTags(enum.IntEnum):
    """List of known / supported tags."""

    TEST_REPORT = 110
    FAILURE_REASON = 112


@enum.unique
class MessageLogType(enum.IntEnum):
    """List of possible received log messages."""

    RESERVED = 0


# Link types and sub-types
type_sub_type_dict = {
    MessageType.COMMAND: MessageCommandType,
    MessageType.REPORT: MessageReportType,
    MessageType.ACK: MessageAckType,
    MessageType.LOG: MessageLogType,
}


class Message:
    """A message who fit testApp protocol.

    The created message is a tlv style message with the following format:
    TYPE: msg_type | message_token | sub_type | errorCode |
    """

    crc_byte_size = 2
    header_size = 8
    max_payload_size = 0xFF
    max_message_size = header_size + max_payload_size + crc_byte_size
    reserved = 0

    def __init__(
        self,
        msg_type: Union[int, MessageType] = 0,
        sub_type: int = 0,
        error_code: int = 0,
        test_suite: int = 0,
        test_case: int = 0,
        tlv_dict: Optional[Dict[str, Union[int, bytes]]] = None,
    ):
        """Create a generic message.

        :param msg_type: Message type
        :type msg_type: MessageType

        :param sub_type: Message sub-type
        :type sub_type: Message<MessageType>Type

        :param error_code: Error value
        :type error_code: integer

        :param test_suite: Suite value
        :type test_suite: integer

        :param test_case: Test value
        :type test_case: integer

        :param tlv_dict: Dictionary containing tlvs elements in the form {'type':'value', ...}
        :type tlv_dict: dict
        """
        self.msg_type = msg_type
        global msg_cnt
        self.msg_token = next(msg_cnt)
        self.sub_type = sub_type
        self.error_code = error_code
        self.test_suite = test_suite
        self.test_case = test_case
        self.tlv_dict = tlv_dict

    def __str__(self):
        """String representation of a message object."""
        string = "msg_type:{}, message_token:{}, type:{}, error_code:{}, reserved:{}, test_suite ID:{}, test_case ID:{}".format(
            self.msg_type,
            self.msg_token,
            self.sub_type,
            self.error_code,
            self.reserved,
            self.test_suite,
            self.test_case,
        )
        if self.tlv_dict is not None:
            # Convert dec to ascii
            tlv = {key: "".join(chr(i) for i in val) for key, val in self.tlv_dict.items()}
            string += f", tlv_dict:{tlv}"
        return string

    def serialize(self) -> bytes:
        """Serialize message into raw packet.

        Format: | msg_type (1b)     | msg_token (1b)  | sub_type (1b)  | error_code (1b)     |
                | test_section (1b) | test_suite (1b) | test_case (1b) | payload_length (1b) |
                | tlv_type (1b)     | tlv_size (1b)   | ...            | crc_checksum (2b)

        :return: bytes representing the Message object
        """
        raw_packet = b""

        raw_packet += struct.pack(
            "BBBBBBB",
            ((int(self.msg_type) << 4) | (1 << 6)),
            self.msg_token,
            int(self.sub_type),
            self.error_code,
            self.reserved,
            self.test_suite,
            self.test_case,
        )

        # Calculate and convert the dictionaries tlv elements into bytes
        if self.tlv_dict is not None:
            payload = b""
            for key, value in self.tlv_dict.items():
                # Check first if it the dict is conform
                parsed_key = b""
                if isinstance(key, TlvKnownTags):
                    parsed_key = struct.pack("B", int(key))
                else:
                    log.internal_warning("{} is not a supported format".format(key))
                parsed_value = b""
                if isinstance(value, str):  # If string given
                    parsed_value = parsed_value.join([struct.pack("B", ord(val)) for val in value])
                elif isinstance(value, int):
                    parsed_value = struct.pack("H", value)  # TODO check endianness later on
                elif isinstance(value, bytes):
                    parsed_value = value
                else:
                    log.internal_warning("{} is not a supported format".format(value))
                # Add the TLV element:
                payload += parsed_key
                payload += struct.pack("B", len(parsed_value))
                payload += parsed_value
            # Add the tlvs elements within the raw_packet
            raw_packet += struct.pack("B", len(payload))
            raw_packet += payload
        else:
            # Add the payload length to 0
            raw_packet += struct.pack("B", 0)

        # Add crc to raw_packet
        raw_packet += struct.pack("H", self.get_crc(raw_packet, self.crc_byte_size))

        return raw_packet

    @classmethod
    def parse_packet(cls, raw_packet: bytes) -> Message:
        """Factory function to create a Message object from raw data.

        :param raw_packet: array of a received message

        :return: itself
        """
        msg = cls()
        if (not isinstance(raw_packet, bytes)) and (len(raw_packet) < (msg.header_size + msg.crc_byte_size)):
            log.error("Packet is not understandable")

        # Check the CRC
        crc = cls.get_crc(raw_packet[: -msg.crc_byte_size], msg.crc_byte_size)
        if crc != struct.unpack("H", raw_packet[-msg.crc_byte_size :])[0]:
            log.error(f"CRC check failed {crc} != {struct.unpack('H', raw_packet[-msg.crc_byte_size:])[0]}")

        unpack_header = struct.unpack("BBBBBBB", raw_packet[:7])

        msg.msg_type = (MessageType)(int((unpack_header[0] & 0x30) >> 4))
        msg.msg_token = int(unpack_header[1])
        # Because the sub-type depend on the type:
        msg.sub_type = (type_sub_type_dict[msg.msg_type])(int(unpack_header[2]))
        msg.error_code = int(unpack_header[3])
        msg.reserved = int(unpack_header[4])
        msg.test_suite = int(unpack_header[5])
        msg.test_case = int(unpack_header[6])
        payload_length = int(raw_packet[7])
        # Create payload based on known tlvs
        if payload_length != 0:
            msg.tlv_dict = {}
            for tag, value in cls._parse_tlv(raw_packet[msg.header_size : -msg.crc_byte_size]):
                msg.tlv_dict[TlvKnownTags(tag)] = value

        return msg

    @classmethod
    def _parse_tlv(cls, tlv_packet: bytes) -> tuple:
        """Generator used to parse TLV formatted bytes array.

        :param tlv_packet: raw TLV formatted bytes array

        :return: tuple containing the extract tag(int) and value(list)
        """
        tlv_iterator = iter(tlv_packet)
        while True:
            try:
                tag = int(next(itertools.islice(tlv_iterator, 1)))
                length = int(next(itertools.islice(tlv_iterator, 1)))
                value = [val for val in itertools.islice(tlv_iterator, length)]
                yield (tag, value)
            except StopIteration:
                break

    def generate_ack_message(self, ack_type: int) -> Union[Message, None]:
        """Generate acknowledgement to send out.

        :param ack_type: ack or nack

        :return: filled acknowledge message otherwise None
        """
        # Return if wrong parameter given
        if not isinstance(ack_type, MessageAckType):
            return None
        # Create ack message
        ack_message = Message(
            msg_type=MessageType.ACK,
            sub_type=ack_type,
            error_code=0,
            test_suite=self.test_suite,
            test_case=self.test_case,
        )
        ack_message.msg_token = self.msg_token
        # Return the ack message
        return ack_message

    def check_if_ack_message_is_matching(self, ack_message: Message) -> bool:
        """Check if the ack message was for this sent message.

        :param ack_message: received acknowledge message

        :return: True if message type and token are valid otherwise
            False
        """
        if ack_message.msg_type == MessageType.ACK and ack_message.msg_token == self.msg_token:
            return True
        else:
            log.internal_info("ack_message: {} \ndifferent of \nthis message: {}".format(str(ack_message), str(self)))
            return False

    def get_message_type(self) -> Union[int, MessageType]:
        """Return actual message type."""
        return self.msg_type

    def get_message_token(self) -> int:
        """Return actual message token."""
        return self.msg_token

    def get_message_sub_type(self) -> int:
        """Return actual message subtype."""
        return self.sub_type

    def get_message_tlv_dict(self) -> dict:
        """Return actual message type/length/value dictionary."""
        return self.tlv_dict

    @classmethod
    def get_crc(cls, serialized_msg: bytes, crc_byte_size: int = 2) -> int:
        """Get the CRC checksum for a bytes message.

        :param serialized_msg: message used for the crc calculation
        :param crc_byte_size: number of bytes dedicated for the crc

        :return: CRC checksum
        """

        crc = 0
        crc_mask = 255
        crc_size = (2 ** (crc_byte_size * 8)) - 1

        for _, msg in enumerate(serialized_msg):
            crc = ((crc >> 8) | (crc << 8)) & crc_size
            crc ^= int(msg)
            crc ^= (crc & crc_mask) >> 4
            crc ^= (crc << 12) & crc_size
            crc ^= ((crc & crc_mask) << 5) & crc_size
        return crc



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
Instrument control auxiliary plugin
***********************************

:module: instrument_control_auxiliary

:synopsis: implementation of existing InstrumentControlAuxiliary for
    Robot framework usage.

.. currentmodule:: instrument_control_auxiliary

"""

from robot.api.deco import keyword, library

from ..auxiliaries.instrument_control_auxiliary import InstrumentControlAuxiliary as InstAux
from .aux_interface import RobotAuxInterface


@library(version="0.1.0")
class InstrumentControlAuxiliary(RobotAuxInterface):
    """Robot framework plugin for InstrumentControlAuxiliary."""

    ROBOT_LIBRARY_SCOPE = "SUITE"

    def __init__(self):
        """Initialize attributes."""
        super().__init__(aux_type=InstAux)

    @keyword(name="Write")
    def write(self, write_command: str, aux_alias: str, validation: tuple = None) -> str:
        """Send a write request to the instrument and then returns if
        the value was successfully written. A query is sent immediately
        after the writing and the answer is compared to the expected
        one.

        :param write_command: write command to send
        :param aux_alias: auxiliary's alias
        :param validation: tuple of the form
            (validation command (str), expected output (str))

        :return: status message depending on the command validation:
            SUCCESS, FAILURE or NO_VALIDATION
        """
        aux = self._get_aux(aux_alias)
        return aux.write(write_command, validation)

    @keyword(name="Read")
    def read(self, aux_alias: str) -> str:
        """Send a read request to the instrument.

        :param aux_alias: auxiliary's alias

        :return: Response message, None if the request expired with a
            timeout.
        """
        aux = self._get_aux(aux_alias)
        return aux.read()

    @keyword(name="Query")
    def query(self, query_command: str, aux_alias: str) -> str:
        """Send a query request to the instrument.

        :param query_command: query command to send
        :param aux_alias: auxiliary's alias

        :return: Response message, None if the request expired with a
            timeout.
        """
        aux = self._get_aux(aux_alias)
        return aux.query(query_command)

    @keyword(name="Get identification")
    def get_identification(self, aux_alias: str) -> str:
        """Get the identification information of an instrument.

        :param aux_alias: auxiliary's alias

        :return: the instrument's identification information
        """
        aux = self._get_aux(aux_alias)
        return aux.helpers.get_identification()

    @keyword(name="Get status byte")
    def get_status_byte(self, aux_alias: str) -> str:
        """Get the status byte of an instrument.

        :param aux_alias: auxiliary's alias

        :return: the intrument's status byte
        """
        aux = self._get_aux(aux_alias)
        return aux.helpers.get_status_byte()

    @keyword(name="Get all errors")
    def get_all_errors(self, aux_alias: str) -> str:
        """Get all errors of an instrument.

        :param aux_alias: auxiliary's alias

        return: list of off errors
        """
        aux = self._get_aux(aux_alias)
        return aux.helpers.get_all_errors()

    @keyword(name="Reset")
    def reset(self, aux_alias: str) -> str:
        """Reset an instrument.

        :param aux_alias: auxiliary's alias

        :return: NO_VALIDATION status code
        """
        aux = self._get_aux(aux_alias)
        return aux.helpers.reset()

    @keyword(name="Self test")
    def self_test(self, aux_alias: str) -> str:
        """Performs a self-test of an instrument.

        :param aux_alias: auxiliary's alias

        :return: the query's response message
        """
        aux = self._get_aux(aux_alias)
        return aux.helpers.self_test()

    @keyword(name="Get remote control state")
    def get_remote_control_state(self, aux_alias: str) -> str:
        """Get the remote control mode (ON or OFF) of an instrument.

        :param aux_alias: auxiliary's alias

        :return: the remote control state
        """
        aux = self._get_aux(aux_alias)
        return aux.helpers.get_remote_control_state()

    @keyword(name="Set remote control on")
    def set_remote_control_on(self, aux_alias: str) -> str:
        """Enables the remote control of an instrument. The instrument
        will respond to all SCPI commands.

        :param aux_alias: auxiliary's alias

        :return: the writing operation's status code
        """
        aux = self._get_aux(aux_alias)
        return aux.helpers.set_remote_control_on()

    @keyword(name="Set remote control off")
    def set_remote_control_off(self, aux_alias: str) -> str:
        """Disable the remote control of an instrument. The instrument
        will respond to query and read commands only.

        :param aux_alias: auxiliary's alias

        :return: the writing operation's status code
        """
        aux = self._get_aux(aux_alias)
        return aux.helpers.set_remote_control_off()

    @keyword(name="Get output channel")
    def get_output_channel(self, aux_alias: str) -> str:
        """Get the currently selected output channel of an instrument.

        :param aux_alias: auxiliary's alias

        :return: the currently selected output channel
        """
        aux = self._get_aux(aux_alias)
        return aux.helpers.get_output_channel()

    @keyword(name="Set output channel")
    def set_output_channel(self, channel: int, aux_alias: str) -> str:
        """Set the output channel of an instrument.

        :param channel: the output channel to select on the instrument
        :param aux_alias: auxiliary's alias

        :return: the writing operation's status code
        """
        aux = self._get_aux(aux_alias)
        return aux.helpers.set_output_channel(channel)

    @keyword(name="Get output state")
    def get_output_state(self, aux_alias: str) -> str:
        """Get the output status (ON or OFF, enabled or disabled) of the
        currently selected channel of an instrument.

        :param aux_alias: auxiliary's alias

        :return: the output state (ON or OFF)
        """
        aux = self._get_aux(aux_alias)
        return aux.helpers.get_output_state()

    @keyword(name="Enable output")
    def enable_output(self, aux_alias: str) -> str:
        """Enable output on the currently selected output channel of an
        instrument.

        :param aux_alias: auxiliary's alias

        :return: the writing operation's status code
        """
        aux = self._get_aux(aux_alias)
        return aux.helpers.enable_output()

    @keyword(name="Disable output")
    def disable_output(self, aux_alias: str) -> str:
        """Disable output on the currently selected output channel of an
        instrument.

        :param aux_alias: auxiliary's alias

        :return: the writing operation's status code
        """
        aux = self._get_aux(aux_alias)
        return aux.helpers.disable_output()

    @keyword(name="Get nominal voltage")
    def get_nominal_voltage(self, aux_alias: str) -> str:
        """Query the nominal voltage of an instrument on the selected
        channel (in V).

        :param aux_alias: auxiliary's alias

        :return: the nominal voltage
        """
        aux = self._get_aux(aux_alias)
        return aux.helpers.get_nominal_voltage()

    @keyword(name="Get nominal current")
    def get_nominal_current(self, aux_alias: str) -> str:
        """Query the nominal current of an instrument on the selected
        channel (in A)

        :param aux_alias: auxiliary's alias

        :return: the nominal current
        """
        aux = self._get_aux(aux_alias)
        return aux.helpers.get_nominal_current()

    @keyword(name="Get nominal power")
    def get_nominal_power(self, aux_alias: str) -> str:
        """Query the nominal power of an instrument on the selected
        channel (in W).

        :param aux_alias: auxiliary's alias

        :return: the nominal power
        """
        aux = self._get_aux(aux_alias)
        return aux.helpers.get_nominal_power()

    @keyword(name="Measure voltage")
    def measure_voltage(self, aux_alias: str) -> str:
        """Return the measured output voltage of an instrument (in V).

        :param aux_alias: auxiliary's alias

        :return: the measured voltage
        """
        aux = self._get_aux(aux_alias)
        return aux.helpers.measure_voltage()

    @keyword(name="Measure current")
    def measure_current(self, aux_alias: str) -> str:
        """Return the measured output current of an instrument (in A).

        :param aux_alias: auxiliary's alias

        :return: the measured current
        """
        aux = self._get_aux(aux_alias)
        return aux.helpers.measure_current()

    @keyword(name="Measure power")
    def measure_power(self, aux_alias: str) -> str:
        """Return the measured output power of an instrument (in W).

        :param aux_alias: auxiliary's alias

        :return: the measured power
        """
        aux = self._get_aux(aux_alias)
        return aux.helpers.measure_power()

    @keyword(name="Get target voltage")
    def get_target_voltage(self, aux_alias: str) -> str:
        """Get the desired output voltage (in V) of an instrument.

        :param aux_alias: auxiliary's alias

        :return: the target voltage
        """
        aux = self._get_aux(aux_alias)
        return aux.helpers.get_target_voltage()

    @keyword(name="Get target current")
    def get_target_current(self, aux_alias: str) -> str:
        """Get the desired output current (in A) of an instrument.

        :param aux_alias: auxiliary's alias

        :return: the target current
        """
        aux = self._get_aux(aux_alias)
        return aux.helpers.get_target_current()

    @keyword(name="Get target power")
    def get_target_power(self, aux_alias: str) -> str:
        """Get the desired output power (in W) of an instrument.

        :param aux_alias: auxiliary's alias

        :return: the target power
        """
        aux = self._get_aux(aux_alias)
        return aux.helpers.get_target_power()

    @keyword(name="Set target voltage")
    def set_target_voltage(self, value: float, aux_alias: str) -> str:
        """Set the desired output voltage (in V) of an instrument.

        :param value: value to be set on the instrument
        :param aux_alias: auxiliary's alias

        :return: the writing operation's status code
        """
        aux = self._get_aux(aux_alias)
        return aux.helpers.set_target_voltage(value)

    @keyword(name="Set target current")
    def set_target_current(self, value: float, aux_alias: str) -> str:
        """Set the desired output current (in A) of an instrument.

        :param value: value to be set on the instrument
        :param aux_alias: auxiliary's alias

        :return: the writing operation's status code
        """
        aux = self._get_aux(aux_alias)
        return aux.helpers.set_target_current(value)

    @keyword(name="Set target power")
    def set_target_power(self, value: float, aux_alias: str) -> str:
        """Set the desired output power (in W) of an instrument.

        :param value: value to be set on the instrument
        :param aux_alias: auxiliary's alias

        :return: the writing operation's status code
        """
        aux = self._get_aux(aux_alias)
        return aux.helpers.set_target_power(value)

    @keyword(name="Get voltage limit low")
    def get_voltage_limit_low(self, aux_alias: str) -> str:
        """Returns the voltage lower limit (in V) of an instrument.

        :param aux_alias: auxiliary's alias

        :return: the query's response message
        """
        aux = self._get_aux(aux_alias)
        return aux.helpers.get_voltage_limit_low()

    @keyword(name="Get voltage limit high")
    def get_voltage_limit_high(self, aux_alias: str) -> str:
        """Returns the voltage upper limit (in V) of an instrument.

        :param aux_alias: auxiliary's alias

        :return: the query's response message
        """
        aux = self._get_aux(aux_alias)
        return aux.helpers.get_voltage_limit_high()

    @keyword(name="Get current limit low")
    def get_current_limit_low(self, aux_alias: str) -> str:
        """Returns the current lower limit (in V) of an instrument.

        :param aux_alias: auxiliary's alias

        :return: the query's response message
        """
        aux = self._get_aux(aux_alias)
        return aux.helpers.get_current_limit_low()

    @keyword(name="Get current limit high")
    def get_current_limit_high(self, aux_alias: str) -> str:
        """Returns the current upper limit (in V) of an instrument.

        :param aux_alias: auxiliary's alias

        :return: the query's response message
        """
        aux = self._get_aux(aux_alias)
        return aux.helpers.get_current_limit_high()

    @keyword(name="Get power limit high")
    def get_power_limit_high(self, aux_alias: str) -> str:
        """Returns the power upper limit (in W) of an instrument.

        :param aux_alias: auxiliary's alias

        :return: the query's response message
        """
        aux = self._get_aux(aux_alias)
        return aux.helpers.get_power_limit_high()

    @keyword(name="Set voltage limit low")
    def set_voltage_limit_low(self, limit_value: float, aux_alias: str) -> str:
        """Set the voltage lower limit (in V) of an instrument.

        :param limit_value: limit value to be set on the instrument
        :param aux_alias: auxiliary's alias

        :return: the writing operation's status code
        """
        aux = self._get_aux(aux_alias)
        return aux.helpers.set_voltage_limit_low(limit_value)

    @keyword(name="Set voltage limit high")
    def set_voltage_limit_high(self, limit_value: float, aux_alias: str) -> str:
        """Set the voltage upper limit (in V) of an instrument.

        :param limit_value: limit value to be set on the instrument
        :param aux_alias: auxiliary's alias

        :return: the writing operation's status code
        """
        aux = self._get_aux(aux_alias)
        return aux.helpers.set_voltage_limit_high(limit_value)

    @keyword(name="Set current limit low")
    def set_current_limit_low(self, limit_value: float, aux_alias: str) -> str:
        """Set the current lower limit (in A) of an instrument.

        :param limit_value: limit value to be set on the instrument
        :param aux_alias: auxiliary's alias

        :return: the writing operation's status code
        """
        aux = self._get_aux(aux_alias)
        return aux.helpers.set_current_limit_low(limit_value)

    @keyword(name="Set current limit high")
    def set_current_limit_high(self, limit_value: float, aux_alias: str) -> str:
        """Set the current upper limit (in A) of an instrument.

        :param limit_value: limit value to be set on the instrument
        :param aux_alias: auxiliary's alias

        :return: the writing operation's status code
        """
        aux = self._get_aux(aux_alias)
        return aux.helpers.set_current_limit_high(limit_value)

    @keyword(name="Set power limit high")
    def set_power_limit_high(self, limit_value: float, aux_alias: str) -> str:
        """Set the power upper limit (in W) of an instrument.

        :param limit_value: limit value to be set on the instrument
        :param aux_alias: auxiliary's alias

        :return: the writing operation's status code
        """
        aux = self._get_aux(aux_alias)
        return aux.helpers.set_power_limit_high(limit_value)



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
Config Registry
***************

:module: config_registry

:synopsis: register auxiliaries and connectors to provide them for import.

.. currentmodule:: config_registry

"""
from __future__ import annotations

from collections import defaultdict
from contextlib import contextmanager
from typing import TYPE_CHECKING, Any, Dict, Iterator, List, Tuple, Type

from ..exceptions import PykisoError
from .dynamic_loader import DynamicImportLinker

if TYPE_CHECKING:
    from pykiso import AuxiliaryInterface
    from pykiso.types import AuxiliaryAlias, AuxiliaryConfig, ConfigDict, ConnectorAlias, ConnectorConfig


class ConfigRegistry:
    """Register auxiliaries with connectors to provide systemwide import
    statements.

    Internally patch the user-configuration if multiple auxiliaries share
    a same communication channel.
    """

    _linker = None

    @staticmethod
    def _make_proxy_channel_config(
        aux_name: AuxiliaryConfig,
    ) -> Tuple[ConnectorAlias, ConnectorConfig]:
        """Craft the configuration dictionary for a proxy communication
        channel to attach to the auxiliary instead of the 'physical' channel.

        :param aux_name: name of the auxiliary to which the proxy channel
            should be plugged.
        :return: the resulting proxy channel name and its configuration as
            as a tuple.
        """
        from pykiso.lib.connectors.cc_proxy import CCProxy

        cchannel_class = CCProxy
        name = f"proxy_channel_{aux_name}"
        config = {
            "config": None,
            "type": f"{cchannel_class.__module__}:{cchannel_class.__name__}",
        }
        return name, config

    @staticmethod
    def _make_proxy_aux_config(
        channel_name: ConnectorAlias,
        aux_list: List[AuxiliaryAlias],
        auto_start: bool,
    ) -> Tuple[AuxiliaryAlias, AuxiliaryConfig]:
        """Craft the configuration dictionary for a proxy auxiliary to be
        attached to all auxiliaries sharing a communication channel.

        :param channel_name: name of the 'physical' channel that should
            be attached to the proxy auxiliary (the only direct connection).
        :return: the resulting proxy auxiliary name and its configuration as
            as a tuple.
        """
        from pykiso.lib.auxiliaries.proxy_auxiliary import ProxyAuxiliary

        aux_class = ProxyAuxiliary
        name = f"proxy_aux_{channel_name}"
        config = {
            "connectors": {"com": channel_name},
            "config": {"aux_list": aux_list, "auto_start": auto_start},
            "type": f"{aux_class.__module__}:{aux_class.__name__}",
        }
        return name, config

    @staticmethod
    def _link_cchannel_to_auxiliaries(
        config: ConfigDict,
    ) -> Dict[ConnectorAlias, List[AuxiliaryAlias]]:
        """Go through each auxiliary configuration, link each found
        channel name to the auxiliary(s) that hold it.

        :param config: dictionary containing yaml configuration content
        :return: a dictionary linking each channel name to the list of
            auxiliaries that are connected to the channel.
        """
        cchannel_to_auxiliaries = defaultdict(list)
        for auxiliary, aux_details in config["auxiliaries"].items():
            try:
                cchannel = aux_details["connectors"]["com"]
            except KeyError:
                # auxiliary doesn't have a communication channel attached
                continue

            cchannel_to_auxiliaries[cchannel].append(auxiliary)
        return cchannel_to_auxiliaries

    @classmethod
    def register_aux_con(cls, config: ConfigDict) -> None:
        """Create import hooks. Register auxiliaries and connectors.

        :param config: dictionary containing yaml configuration content
        """
        # 1. Detect required proxy setups
        cchannel_to_auxiliaries = cls._link_cchannel_to_auxiliaries(config)
        proxies = []

        # 2. Overwrite auxiliary and connector config with required proxies
        for channel_name, auxiliaries in cchannel_to_auxiliaries.items():
            if len(auxiliaries) < 2:
                # only one auxiliary holds the channel so no proxy is required
                continue

            # automatically start proxy if at least one auxiliary has the auto_start flag set
            for auxiliary in auxiliaries:
                try:
                    auto_start = config["auxiliaries"][auxiliary]["config"]["auto_start"]
                except (KeyError, TypeError):
                    # default value for auto_start is True
                    auto_start = True
                if auto_start:
                    break

            # create a proxy auxiliary config for this shared channel
            proxy_aux_name, proxy_aux_cfg = cls._make_proxy_aux_config(channel_name, auxiliaries, auto_start)
            config["auxiliaries"][proxy_aux_name] = proxy_aux_cfg
            proxies.append(proxy_aux_name)

            # create a proxy channel config for each of the auxiliaries sharing the channel
            for aux_name in auxiliaries:
                cc_proxy_name, cc_proxy_cfg = cls._make_proxy_channel_config(aux_name)
                config["auxiliaries"][aux_name]["connectors"]["com"] = cc_proxy_name
                config["connectors"][cc_proxy_name] = cc_proxy_cfg

        # 3. Create and install the auxiliary import hook
        cls._linker = DynamicImportLinker()
        cls._linker.install()

        # 4. Provide auxiliaries and connectors from potentially patched configuration
        for connector, con_details in config["connectors"].items():
            cfg = con_details.get("config") or dict()
            cls._linker.provide_connector(connector, con_details["type"], **cfg)

        for auxiliary, aux_details in config["auxiliaries"].items():
            cfg = aux_details.get("config") or dict()
            cls._linker.provide_auxiliary(
                auxiliary,
                aux_details["type"],
                aux_cons=aux_details.get("connectors") or dict(),
                **cfg,
            )

        # 5. Finally, import required ProxyAuxiliary instances so that user doesn't have to
        for proxy_aux in proxies:
            try:
                cls.get_aux_by_alias(proxy_aux)
            except PykisoError:
                # ensure that the created ProxyAuxiliary is stopped if its creation fails
                cls.delete_aux_con()
                raise

    @classmethod
    def delete_aux_con(cls) -> None:
        """Deregister the import hooks, close all running threads,
        delete all instances.
        """
        cls._linker.uninstall()
        cls._linker = None

    @classmethod
    @contextmanager
    def provide_auxiliaries(cls, config: ConfigDict) -> Iterator[None]:
        """Context manager that registers importable auxiliary
        aliases and cleans them up at exit.

        :param config: config dictionary from the YAML configuration
            file.

        :yield: None
        """
        try:
            cls.register_aux_con(config)
            yield
        finally:
            cls.delete_aux_con()

    @classmethod
    def get_all_auxes(cls) -> Dict[AuxiliaryAlias, AuxiliaryInterface]:
        """Return all auxiliaires instances and alias

        :return: dictionary with alias as keys and instances as values
        """
        return cls._linker._aux_cache.instances

    @classmethod
    def get_auxes_by_type(cls, aux_type: Type[AuxiliaryInterface]) -> Dict[str, AuxiliaryInterface]:
        """Return all auxiliaries who match a specific type.

        :param aux_type: auxiliary class type (DUTAuxiliary,
            CommunicationAuxiliary...)

        :return: dictionary with alias as keys and instances as values
        """
        all_auxes = cls._linker._aux_cache.instances
        return {alias: inst for alias, inst in all_auxes.items() if isinstance(inst, aux_type)}

    @classmethod
    def get_aux_by_alias(cls, alias: AuxiliaryAlias) -> AuxiliaryInterface:
        """Return the associated auxiliary instance to the given alias.

        :param alias: auxiliary's alias
        :return: auxiliary instance created by the dymanic loader
        """
        return cls._linker._aux_cache.get_instance(alias)

    @classmethod
    def get_aux_config(cls, name: AuxiliaryAlias) -> Dict[str, Any]:
        """Return the registered auxiliary configuration based on his
        name.

        :param name: auxiliary's alias
        :return: auxiliary's configuration (yaml content)
        """
        return cls._linker._aux_cache.configs[name]

    @classmethod
    def get_auxes_alias(cls) -> List[AuxiliaryAlias]:
        """return all created auxiliaries alias.

        :return: list containing all auxiliaries alias
        """
        return [alias for alias in cls._linker._aux_cache.connectors.keys()]



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
Dynamic Loader plugin
*********************

:module: loader

:synopsis: implementation of existing magic import mechanism from ITF
    for Robot framework usage.

.. currentmodule:: loader

"""

import importlib
import sys
from pathlib import Path
from typing import List

from robot.api import logger
from robot.api.deco import keyword, library

from ...cli import parse_config
from ...test_setup.config_registry import ConfigRegistry
from ...test_setup.dynamic_loader import PACKAGE
from ..auxiliaries.proxy_auxiliary import ProxyAuxiliary


@library(version="0.1.0")
class RobotLoader:
    """Robot framework plugin for ITF magic import mechanism."""

    ROBOT_LIBRARY_SCOPE = "GLOBAL"

    def __init__(self, config_file: Path):
        """Initialize attributes.

        :param config_file : yaml configuration file path
        """
        self.config_file = parse_config(config_file)
        self.auxiliaries = {}

    def _detect_proxy_auxes(self) -> List[str]:
        """Detect all proxy auxiliaires from yaml configuration file.

        :return: list containing all proxy auxiliairies alias
        """
        proxy_aux_alias = []
        for alias, aux_conf in self.config_file["auxiliaries"].items():
            if ProxyAuxiliary.__name__ in aux_conf["type"]:
                proxy_aux_alias.append(alias)
        return proxy_aux_alias

    @keyword(name="install")
    def install(self) -> None:
        """Provide, create and import auxiliaires/connectors present
        within yaml configuration file.

        :raises: re-raise the caught exception (Exception level)
        """
        try:
            ConfigRegistry.register_aux_con(self.config_file)
            proxy_auxes = self._detect_proxy_auxes()
            auxes_alias = ConfigRegistry.get_auxes_alias()

            # import all auxiliaries except proxy ones
            for alias in auxes_alias:
                if alias not in proxy_auxes:
                    self.auxiliaries[alias] = importlib.import_module(f".{alias}", f"{PACKAGE}.auxiliaries")
            # finally install proxy auxiliaries
            for alias in proxy_auxes:
                self.auxiliaries[alias] = importlib.import_module(f".{alias}", f"{PACKAGE}.auxiliaries")
            logger.info(f"auxiliaries {self.auxiliaries} installed")
        except Exception as error:
            logger.error(f"An error occurred during auxiliaries creation, reason : {error}")
            raise

    @keyword(name="uninstall")
    def uninstall(self) -> None:
        """Uninstall all created instances of auxiliaires/connectors.

        :raises: re-raise the caught exception (Exception level)
        """
        try:
            ConfigRegistry.delete_aux_con()
            for alias in self.auxiliaries:
                sys.modules.pop(f"{PACKAGE}.auxiliaries.{alias}")
            logger.info(f"auxiliaries {self.auxiliaries} uninstalled")
        except Exception as error:
            logger.error(f"An error occurred during auxiliaries deletion, reason : {error}")
            raise



##########################################################################
# Copyright (c) 2024-2024 Accenture GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
Endpoint Management
*******************

:module: endpoint

:synopsis: Expose pykiso setup as a server

.. currentmodule:: endpoint

"""

import logging

log = logging.getLogger(__name__)

from .auxiliary import AuxiliaryInterface
from .test_setup.config_registry import ConfigRegistry


class Endpoint:
    """Parent class to use for exposing pykiso setup as a server."""

    def __init__(self):
        # Get the list of auxiliaries
        self.auxiliaries: dict[str, AuxiliaryInterface] = ConfigRegistry.get_all_auxes()
        # Expose each auxiliary
        for alias, auxiliary in self.auxiliaries.items():
            self.expose_one_auxiliary(alias, auxiliary)
        # Expose pytkiso common interface
        self.expose_pykiso()
        # Wait for completion or termination
        self.wait_for_termination()

    def expose_one_auxiliary(self, alias: str, auxiliary: AuxiliaryInterface):
        """Method that will be called to expose one auxiliary.

        :param alias: Given name to the auxiliary in the yaml file
        :param auxiliary: auxiliary instance associated with the alias

        :raises NotImplementedError: This method should be implemented in the child class
        """
        raise NotImplementedError

    def expose_pykiso(self):
        """Method that will be called to expose the pykiso common interface.

        .. warning::
            This method will be implemented in the future.
        """
        pass

    def wait_for_termination(self):
        """Blocking method that will wait for the termination of the server.

        :raises NotImplementedError: This method should be implemented in the child class
        """
        raise NotImplementedError



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
uds_auxiliary
*************

:module: uds_auxiliary

:synopsis: Auxiliary used to handle Unified Diagnostic Service protocol

.. currentmodule:: uds_auxiliary

"""
import logging
import threading
import time
from contextlib import contextmanager
from pathlib import Path
from typing import Iterator, List, Optional, Union

try:
    import can
    from uds import IsoServices
except ImportError as e:
    raise ImportError(f"{e.name} dependency missing, consider installing pykiso with 'pip install pykiso[can]'")

from pykiso.connector import CChannel

from .common import uds_exceptions
from .common.uds_base_auxiliary import UdsBaseAuxiliary
from .common.uds_request import UDSCommands
from .common.uds_response import UdsResponse
from .common.uds_utils import get_uds_service

log = logging.getLogger(__name__)


class UdsAuxiliary(UdsBaseAuxiliary):
    """Auxiliary used to handle the UDS protocol on client (tester) side."""

    errors = uds_exceptions

    def __init__(
        self,
        com: CChannel,
        config_ini_path: Optional[Union[Path, str]] = None,
        odx_file_path: Optional[Union[Path, str]] = None,
        request_id: Optional[int] = None,
        response_id: Optional[int] = None,
        tp_layer: dict = None,
        uds_layer: dict = None,
        **kwargs,
    ):
        """Initialize attributes.
        :param com: communication channel connector.
        :param config_ini_path: UDS parameter file.
        :param odx_file_path: ecu diagnostic definition file.
        :param request_id: optional CAN ID used for sending messages.
        :param response_id: optional CAN ID used for receiving messages.
        :param tp_layer: isotp configuration given at yaml level
        :param uds_layer: uds configuration given at yaml level
        """
        self.is_tester_present = None
        super().__init__(
            com,
            config_ini_path,
            odx_file_path,
            request_id,
            response_id,
            tp_layer,
            uds_layer,
            **kwargs,
        )

    def transmit(self, data: bytes, req_id: int, extended: bool = False) -> None:
        """Transmit a message through ITF connector. This method is a
        substitute to transmit method present in python-uds package.

        :param data: data to send
        :param req_id: CAN message identifier
        :param extended: True if addressing mode is extended otherwise
            False
        """
        self.channel._cc_send(msg=data, remote_id=req_id)

    def send_uds_raw(
        self,
        msg_to_send: Union[bytes, List[int], tuple],
        timeout_in_s: float = 6,
        response_required: bool = True,
    ) -> Union[UdsResponse, bool]:
        """Send a UDS diagnostic request to the target ECU and check response.

        :param msg_to_send: can uds raw bytes to be sent
        :param timeout_in_s: not used, actual timeout in seconds for the response can be
            configured with the uds_aux.config.uds_layer.p2_can_client parameter
            inside the yaml config file. (default value is 5s)
        :param response_required: Wait for a response if True

        :raise ResponseNotReceivedError: raised when no answer has been received
        :raise Exception: raised when the raw message could not be send properly

        :return: the raw uds response's bytes, or True if a response is
            not expected and the command is properly sent otherwise
            False
        """
        try:
            if log.isEnabledFor(logging.getLogger().level):
                log.internal_info(
                    "UDS request to send '%s'",
                    ["0x{:02X}".format(i) for i in msg_to_send],
                )
            resp = self.uds_config.send(
                msg_to_send,
                responseRequired=response_required,
                tpWaitTime=self.tp_waiting_time,
            )
        except Exception:
            log.exception("Error while sending uds raw request")
            return False

        if resp is None:
            if not response_required:
                return True
            else:
                raise self.errors.ResponseNotReceivedError(msg_to_send)

        resp = UdsResponse(
            resp,
            resp_time=self.uds_config.last_resp_time,
            pending_resp_times=self.uds_config.last_pending_resp_times,
        )
        log.internal_info("UDS response received %s", resp)
        return resp

    @staticmethod
    def check_max_pending_time(resp: UdsResponse, max_pending_time: float) -> bool:
        """Check that the time between pending response messages does not exceed a
        max value

        :param resp: uds response to check
        :param max_pending_time: max pending time, in seconds

        :return: True if the max value is not exceeded, False otherwise
        """
        return max_pending_time >= max(resp.pending_resp_times)

    def check_raw_response_positive(self, resp: UdsResponse) -> Optional[bool]:
        """Check if the response is positive, raise an error if not

        :param resp: raw response of uds request

        :raise UnexpectedResponseError: raised when the answer is not the expected one

        :return: True if response is positive
        """
        if resp.is_negative:
            log.internal_info(f"Negative response with NRC: {resp.nrc.name}")
            raise self.errors.UnexpectedResponseError(resp)
        return True

    def check_raw_response_negative(self, resp: UdsResponse) -> Optional[bool]:
        """Check if the response is negative, raise an error if not

        :param resp: raw response of uds request

        :raise UnexpectedResponseError: raised when the answer is not the expected one

        :return: True if response is negative
        """
        if not resp.is_negative:
            raise self.errors.UnexpectedResponseError(resp)
        log.internal_info("Negative response with : %s", resp.nrc.name)
        return True

    def send_uds_config(
        self,
        msg_to_send: dict,
        timeout_in_s: float = 6,
    ) -> Union[dict, bool]:
        """Send UDS config to the target ECU.

        :param msg_to_send: uds config to be sent
        :param timeout_in_s: not used

        :return: a dict containing the uds response, or True if a
            response is not expected and the command is properly sent
            otherwise False
        """
        if not self.uds_config_enable:
            log.error("Uds configured without ODX, sending Uds config impossible!")
            return False

        try:
            uds_service_name = get_uds_service(msg_to_send["service"])
            uds_service = getattr(self.uds_config, uds_service_name)
            req_resp_data = uds_service(**msg_to_send["data"])
            if req_resp_data is None:
                req_resp_data = True
            log.internal_info("UDS response received %s", req_resp_data)
            return req_resp_data
        except AttributeError:
            # Service not found, raised by getattr()
            log.exception(f"Could not send UDS config request: unknown service {uds_service_name}")
            return False
        except Exception:
            # Data send failed
            log.exception("Could not send UDS config request: sending failed")
            return False

    def hard_reset(self) -> Union[dict, UdsResponse]:
        """Allow power reset of the component

        :return: response of the hard reset request
        """
        return self.send_uds_raw(UDSCommands.ECUReset.HARD_RESET)

    def force_ecu_reset(self) -> UdsResponse:
        """Allow power reset of the component

        :return: response of the force ecu reset request
        """
        if not self.uds_config_enable:
            return self.send_uds_raw(UDSCommands.ECUReset.KEY_OFF_ON_RESET)
        else:
            ecu_hard_reset_req = {
                "service": IsoServices.EcuReset,
                "data": {"parameter": "ForceEcuReset"},
            }
            return self.send_uds_config(ecu_hard_reset_req)

    def soft_reset(self) -> Union[dict, UdsResponse]:
        """Perform soft reset of the component, equivalent to a restart of application

        :return: response of the soft reset request
        """
        if not self.uds_config_enable:
            return self.send_uds_raw(UDSCommands.ECUReset.SOFT_RESET)
        else:
            ecu_soft_reset_req = {
                "service": IsoServices.EcuReset,
                "data": {"parameter": "Soft Reset"},
            }
            return self.send_uds_config(ecu_soft_reset_req)

    def read_data(self, parameter: str) -> Optional[Union[dict, bool]]:
        """UDS config command that allow data reading

        :param parameter: data to be read

        :return: a dict with uds config response
        """
        if self.uds_config_enable:
            req = {
                "service": IsoServices.ReadDataByIdentifier,
                "data": {
                    "parameter": parameter,
                },
            }
            return self.send_uds_config(req)
        else:
            log.error("No uds config found")
            return

    def write_data(self, parameter: str, value: Union[List[bytes], bytes]) -> Optional[Union[dict, bool]]:
        """UDS config command that allow data writing

        :param parameter: data to be set
        :param value: new content of the data

        :return: a dict with uds config response
        """
        if self.uds_config_enable:
            req = {
                "service": IsoServices.WriteDataByIdentifier,
                "data": {"parameter": parameter, "dataRecord": value},
            }
            return self.send_uds_config(req)
        else:
            log.error("No uds config found")
            return

    def _sender_run(self, period: int, stop_event: threading.Event) -> None:
        """send tester present at defined period until stopped

        :param period: period in seconds to use for the cyclic sending of tester present
        :param stop_event: event to set to stop the sending of tester present
        """
        while not stop_event.is_set():
            self.send_uds_raw(
                UDSCommands.TesterPresent.TESTER_PRESENT_NO_RESPONSE,
                response_required=False,
            )
            stop_event.wait(period)

    @contextmanager
    def tester_present_sender(self, period: int = 4) -> Iterator[None]:
        """Context manager that continuously sends tester present messages via UDS

        :param period: period in seconds to use for the cyclic sending of tester present
        """
        stop_event = threading.Event()
        sender = threading.Thread(
            name="TesterPresentSender",
            target=self._sender_run,
            args=(period, stop_event),
        )
        try:
            yield sender.start()
        finally:
            stop_event.set()
            sender.join()

    def start_tester_present_sender(self, period: int = 4):
        """Start to continuously send tester present messages via UDS"""
        if not self.is_tester_present:
            self.is_tester_present = self.tester_present_sender(period=period)
            return self.is_tester_present.__enter__()

    def stop_tester_present_sender(self):
        """Stop to continuously send tester present messages via UDS"""
        if self.is_tester_present:
            self.is_tester_present.__exit__(None, None, None)
            self.is_tester_present = None
        else:
            log.internal_warning("Tester present sender should be started before it can be stopped")

    def _receive_message(self, timeout_in_s: float) -> None:
        """This method is only used to populate the python-uds reception
        buffer. When a message is received, invoke python-uds configured
        callback to make it available for pyton-uds

        :param timeout_in_s: timeout on reception.
        """

        recv_response = self.channel.cc_receive(timeout=timeout_in_s)
        received_data = recv_response.get("msg")
        arbitration_id = recv_response.get("remote_id")

        if received_data is not None and arbitration_id == self.res_id:
            can_msg = can.Message(
                data=received_data,
                arbitration_id=arbitration_id,
                is_extended_id=False,
            )
            self.uds_config.tp.callback_onReceive(can_msg)

    def _run_command(self, cmd_message, cmd_data=None) -> Union[dict, bytes, bool]:
        """Not used."""
        pass

    def _delete_auxiliary_instance(self) -> bool:
        """Close current associated channel.

        :return: always True
        """
        if self.is_tester_present:
            self.stop_tester_present_sender()
        return super()._delete_auxiliary_instance()



##########################################################################
# Copyright (c) 2010-2023 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
:module: utils

:synopsis: implements common utilities to support the pytest-pykiso duality.
"""

from __future__ import annotations

import sys
import unittest
from typing import Callable, TypeVar

from _pytest.unittest import TestCaseFunction

from pykiso.test_coordinator.test_case import BasicTest
from pykiso.test_coordinator.test_suite import BaseTestSuite

T = TypeVar("T", bound=Callable)


def export(func: T) -> T:
    """Decorator to add a function or a class to its module's ``__all__``."""
    mod = sys.modules[func.__module__]
    if hasattr(mod, "__all__"):
        mod.__all__.append(func.__name__)
    else:
        mod.__all__ = [func.__name__]
    return func


export(export)


@export
def get_base_testcase(item: TestCaseFunction) -> unittest.TestCase:
    """Returns the unittest test case wrapped within a pytest test case.

    :param item: pytest-collected test case.
    :return: wrapped unittest test case (or pykiso test case).
    """
    return item.parent._obj(item.name)


@export
def is_kiso_testcase(tc):
    """
    Determine whether the provided test case is a pykiso test case
    wrapped by pytest.
    """
    return isinstance(tc, TestCaseFunction) and isinstance(get_base_testcase(tc), (BasicTest, BaseTestSuite))



##########################################################################
# Copyright (c) 2010-2023 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
Logging configuration
*********************

:module: logging

:synopsis: initialize the pykiso loggers and patch then onto pytest's loggers
    in order to control CLI and file logging with pytest's built-in options.

"""

from __future__ import annotations

import logging
import os
from typing import TYPE_CHECKING

import pytest
from _pytest.logging import ColoredLevelFormatter

from pykiso.logging_initializer import initialize_logging

from .utils import *

if TYPE_CHECKING:
    from _pytest.logging import LoggingPlugin
    from _pytest.main import Session


@export
@pytest.hookimpl(trylast=True)
def pytest_sessionstart(session: Session):
    """Initialize pykiso logging and patch the resulting loggers onto
    pytest's logger at session start.

    Try to run this hook implementation at last to ensure that logging
    is already preconfigured by pytest.

    :param session: the current test session.
    """
    # get pytest's logger as plugin
    pytest_logger: LoggingPlugin = session.config.pluginmanager.get_plugin("logging-plugin")
    if pytest_logger.log_cli_level is None:
        pytest_logger.log_cli_level = logging.INFO

    # run pykiso's logging initialization
    if pytest_logger.log_file_handler.baseFilename == os.devnull:
        log_file_path = None
    else:
        log_file_path = pytest_logger.log_file_handler.baseFilename

    initialize_logging(
        log_path=log_file_path,
        log_level=logging.getLevelName(pytest_logger.log_cli_level),
        report_type="text",
        # display internal logs at least -vv is provided
        verbose=(session.config.getoption("verbose") > 1),
    )
    root_logger = logging.getLogger()
    # get all handlers that were configured by pykiso to retrieve their level
    stream_handler = next(
        (hdlr for hdlr in root_logger.handlers if type(hdlr) == logging.StreamHandler),
        None,
    )
    file_handler = next(
        (hdlr for hdlr in root_logger.handlers if type(hdlr) == logging.FileHandler),
        None,
    )
    if pytest_logger.log_cli_handler.level != logging.NOTSET and stream_handler is not None:
        pytest_logger.log_cli_level = stream_handler.level
        pytest_logger.log_cli_handler.level = stream_handler.level
        if isinstance(pytest_logger.log_cli_handler.formatter, ColoredLevelFormatter):
            pytest_logger.log_cli_handler.formatter.add_color_level(logging.INTERNAL_WARNING, "yellow", "light")
            pytest_logger.log_cli_handler.formatter.add_color_level(logging.INTERNAL_INFO, "green", "light")
            pytest_logger.log_cli_handler.formatter.add_color_level(logging.INTERNAL_DEBUG, "purple", "light")

    if pytest_logger.log_file_handler.level != logging.NOTSET and file_handler is not None:
        pytest_logger.log_file_handler.level = file_handler.level
        pytest_logger.log_file_level = file_handler.level

    root_logger.handlers.clear()
    pytest_logger.log_level = root_logger.level

    root_logger.addHandler(pytest_logger.log_cli_handler)
    root_logger.addHandler(pytest_logger.log_file_handler)



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
Record Auxiliary
****************

:module: record_auxiliary

:synopsis: Auxiliary used to record a connectors receive channel.

.. currentmodule:: record_auxiliary
"""

import io
import logging
import re
import sys
import threading
import time
from pathlib import Path
from typing import Any, Dict, List, Optional

from pykiso import CChannel
from pykiso.auxiliary import AuxiliaryInterface, close_connector, open_connector
from pykiso.lib.connectors.cc_proxy import CCProxy

log = logging.getLogger(__name__)


class StringIOHandler(io.StringIO):
    def __init__(self, multiprocess: bool = False) -> None:
        """Constructor"""
        super(StringIOHandler, self).__init__()
        self.data_lock = threading.Lock()

    def get_data(self) -> str:
        """Get data from the string

        :return: data from the string
        """
        with self.data_lock:
            return self.getvalue()

    def set_data(self, data: str) -> None:
        """Add data to the already existing data string

        :param data: the data to be write over the existing string
        """
        with self.data_lock:
            self.write(data)


class RecordAuxiliary(AuxiliaryInterface):
    """Auxiliary used to record a connectors receive channel."""

    LOG_HEADER = "Received data :"

    def __init__(
        self,
        com: CChannel,
        is_active: bool = False,
        timeout: float = 0,
        log_folder_path: str = "",
        max_file_size: int = int(5e7),
        multiprocess: bool = False,
        manual_start_record: bool = False,
        **kwargs,
    ) -> None:
        """Constructor.

        :param com: Communication connector to record
        :param is_active: Flag to actively poll receive channel in
            another thread
        :param timeout: timeout for the receive channel
        :param log_path: path to the log folder
        :param max_file_size: maximal size of the data string
        :param multiprocess: deprecated, will not be taken into account.
        :param manual_start_record: flag to not start recording on
            auxiliary creation
        """
        super().__init__(is_proxy_capable=True, tx_task_on=False, rx_task_on=False, **kwargs)
        self.channel = com
        self.is_active = is_active
        self.timeout = timeout
        self.stop_receive_event = None
        self._receive_thread_or_process = None
        self.cursor = 0
        self.log_folder_path = log_folder_path
        self.multiprocess = multiprocess
        self._data = StringIOHandler(multiprocess)
        self.max_file_size = max_file_size

        if self.is_active and not manual_start_record:
            self.start_recording()

    def get_data(self) -> str:
        """Return the entire log buffer content.

        :return: buffer content
        """
        data = self._data.get_data()
        return data

    def set_data(self, data: str) -> None:
        """Add data to the already existing data string.

        :param data: the data to be write over the existing string
        """
        self._data.set_data(data)

    def _create_auxiliary_instance(self) -> bool:
        """Open the connector and start running receive thread
        if is_active is set.

        :return: True if successful
        """

        log.internal_info("Create auxiliary instance")
        log.internal_info("Enable channel")

        try:
            if not self.is_active:
                self.channel.open()
        except Exception:
            log.exception("Error encountered during channel creation.")
            return False
        return True

    def _delete_auxiliary_instance(self) -> bool:
        """Close connector and stop receive thread when is_active flag
        is set.

        :return: always True
        """
        log.internal_info("Delete auxiliary instance")

        self.stop_recording()

        try:
            if not self.is_active:
                self.channel.close()
        except Exception:
            log.exception("Unable to close Channel.")

        return True

    def receive(self) -> None:
        """Open channel and actively poll the connectors receive channel.
        Stop and close connector when stop receive event has been set.
        """
        try:
            self.channel.open()
        except Exception:
            log.exception("Error encountered while channel creation.")
            return

        log.internal_info(f"Received message/log at {self.log_folder_path}")
        self._data.write(self.LOG_HEADER)
        while not self.stop_receive_event.is_set():
            if sys.getsizeof(self.get_data()) > self.max_file_size:
                log.error("Data size too large")

            recv_response = self.channel.cc_receive(timeout=self.timeout)

            stream = recv_response.get("msg")
            source = recv_response.get("remote_id")

            if stream:
                stream = self.parse_bytes(stream)
                if source is not None:
                    self.set_data(f"\n{source}    {stream}")
                else:
                    self.set_data("\n" + stream)

        try:
            self.channel.close()
        except Exception:
            log.exception("Error encountered while closing channel.")

    @staticmethod
    def parse_bytes(data: bytes) -> str:
        """Decode the received bytes

        :param data: data to be decoded

        :return: data decoded
        """
        # decode
        try:
            return data.decode()
        except (UnicodeDecodeError, AttributeError):
            pass
        # get Hex values
        try:
            return data.hex()
        except AttributeError:
            pass

        # Fail decoding, parse RAW data in string (avoid losing it)
        log.error(f"Could not parse the received data: {data}")
        return str(data)

    def clear_buffer(self) -> None:
        """Clean the buffer that contain received messages."""
        log.internal_info("Clearing buffer")
        self._data = StringIOHandler()
        self.cursor = 0

    def stop_recording(self) -> None:
        """Stop recording."""
        if self._receive_thread_or_process is not None and self._receive_thread_or_process.is_alive():
            self.stop_receive_event.set()
            self._receive_thread_or_process.join()
            log.internal_info(f"{self.name} Recording has stopped")
        else:
            log.internal_info("Already Stopped")

    def start_recording(self) -> None:
        """Clear buffer and start recording."""
        # Ensure no record is on-going
        if self._receive_thread_or_process is None or not self._receive_thread_or_process.is_alive():
            # define Thread variables
            self.stop_receive_event = threading.Event()
            self._receive_thread_or_process = threading.Thread(target=self.receive)

            self.clear_buffer()
            self._receive_thread_or_process.start()
            log.internal_info(f"{self.name} Recording has started")
        else:
            log.internal_info(f"{self.name} Already started")

    def is_log_empty(self) -> bool:
        """Check if logs are available in the log buffer.

        :return: True if log is empty, False either
        """
        return len(self.get_data()) - len(self.LOG_HEADER) <= 0

    def dump_to_file(self, filename: str, mode: str = "w+", data: str = None) -> bool:
        """Writing data in file.

        :param filename: name of the file where data are saved
        :param mode: modes of opening a file (eg: w, a)
        :param data: Optional write/append specific data to the file.

        :return: True if the dumping has been successful, False else

        :raises FileNotFoundError: if the given folder path is not a
            folder
        """
        # check if there are data
        if (data is None and self.is_log_empty()) or data == "":
            log.internal_warning("Log data is empty. skip dump to file.")
            return False

        path_to_file = Path(self.log_folder_path) / filename
        path_to_file.parent.mkdir(parents=True, exist_ok=True)

        with open(path_to_file, mode) as f:
            f.write(data or self.get_data())
            log.internal_info(f"Log written in {path_to_file}.")

        return True

    def search_regex_in_folder(self, regex: str) -> Optional[Dict[str, List[str]]]:
        """Returns all occurrences found by the regex in the logs and
        message received.

        :param regex: str regex to compare to logs

        :return: dictionary with filename and the list of matches with
            regular expression

        :raises FileNotFoundError: if the given folder path is not a
            folder
        """

        regex_in_folder = {}
        log_folder_path = Path(self.log_folder_path)
        if not log_folder_path.is_dir():
            log.error(f"folder {self.log_folder_path} does not exist")
            raise FileNotFoundError(f"Path {log_folder_path} does not exist.")

        for file in log_folder_path.iterdir():
            file_path = log_folder_path / file
            log_file_content = file_path.read_text()
            list_regex_in_file = re.findall(regex, log_file_content, re.MULTILINE)
            regex_in_folder[str(file)] = list_regex_in_file

        return regex_in_folder

    def search_regex_in_file(self, regex: str, filename: str) -> Optional[List[str]]:
        """Returns all occurrences found by the regex in the logs and
        message received.

        :param regex: str regex to compare to logs
        :param filename: filename of the desired file

        :return: list of matches with regular expression in the chosen
            file
        """
        path = Path(self.log_folder_path) / filename
        if not path.exists():
            log.error(f"No such file {path}")
            return None

        log_file_content = path.read_text()
        list_regex_in_file = re.findall(regex, log_file_content, re.MULTILINE)

        return list_regex_in_file

    def search_regex_current_string(self, regex: str) -> Optional[List[str]]:
        """Returns all occurrences found by the regex in the logs and
            message received.

        :param regex: str regex to compare to logs

        :return: list of matches with regular expression in the current
            string
        """
        list_regex_in_string = re.findall(regex, self.get_data(), re.MULTILINE)
        return list_regex_in_string

    def _log_query(
        self,
        from_cursor: bool = True,
        set_cursor: bool = True,
        display_log: bool = False,
    ) -> str:
        """Provide the internal log interaction mechanism.

        :param from_cursor: whether to get the logs from the last cursor
            position (True) or the full logs
        :param set_cursor: whether to update the cursor to the last
            position of the string
        :param display_log: whether to log (via logging) the retrieved
            part or just return it

        :return: string with concerned log(s)
        """
        output = self.get_data()
        end = len(output)
        start = min(self.cursor, end)
        if from_cursor:
            output = output[start:]
        if set_cursor:
            self.cursor = end
        if display_log and output:
            logging.info(output)
        return output

    def previous_log(self) -> str:
        """set cursor position to current position.

        This will also display the logs from the last cursor position
        in the log.

        :return: log from the last current position
        """
        return self._log_query(from_cursor=True, set_cursor=False, display_log=False)

    def new_log(self) -> str:
        """Get new entries (after cursor position) from the log. This will set the cursor.

        :return: return log after cursor
        """
        return self._log_query(from_cursor=True, set_cursor=True, display_log=False)

    def is_message_in_log(
        self,
        message: str,
        from_cursor: bool = True,
        set_cursor: bool = True,
        display_log: bool = False,
    ) -> bool:
        """Check for a message being in log.

        :param message: str message to check presence in logs.
        :param from_cursor: whether to get the logs from the last cursor
            position (True) or the full logs
        :param set_cursor: whether to update the cursor
        :param display_log: whether to log (via logging) the retrieved
            part or just return it

        :return: True if a message is in log, False otherwise.
        """
        ret_logs = self._log_query(
            from_cursor=from_cursor,
            set_cursor=set_cursor,
            display_log=display_log,
        )

        return message in ret_logs

    def is_message_in_full_log(self, message: str):
        """Check for a message being in log.

        :param message: message to check presence in logs.
        :return: True if a message is in log, False otherwise
        """
        ret_logs = self.get_data()
        return message in ret_logs

    def wait_for_message_in_log(
        self,
        message: str,
        timeout: float = 10.0,
        interval: float = 0.1,
        from_cursor: bool = True,
        set_cursor: bool = True,
        display_log: bool = False,
        exception_on_failure: bool = True,
    ) -> bool:
        """Poll log at every interval time, fail if messages has not
        shown up within the specified timeout and exception set to True,
        log an error otherwise.

        :param message: str message expected to show up
        :param timeout: int timeout in seconds for the check
        :param interval: int period in seconds for the log poll
        :param from_cursor: whether to get the logs from the last cursor
            position (True) or the full logs
        :param set_cursor: whether to update the cursor to the last log
            position
        :param display_log: whether to log (via logging) the retrieved
            part or just return it
        :param exception_on_failure: if set, raise a TimeoutError if the
            expected messages wasn't found in the logs. Otherwise,
            simply output a warning.

        :return: True if the message have been received in the log,
            False otherwise

        :raises TimeoutError: when a given message has not arrived in
            time
        """
        start = time.time()
        while not self.is_message_in_log(
            message,
            from_cursor=from_cursor,
            set_cursor=set_cursor,
            display_log=display_log,
        ):
            elapsed_time = time.time() - start
            if elapsed_time > timeout:
                if exception_on_failure:
                    raise TimeoutError(
                        f"Maximum wait time for message {message} " f"in log exceeded (waited {elapsed_time:.1f}s)."
                    )
                else:
                    logging.warning(
                        f"Maximum wait time for message {message} " f"in log exceeded (waited {elapsed_time:.1f}s)."
                    )
                    return False
            time.sleep(interval)
        logging.info(f"Received message after {(time.time() - start):.1f}s")
        return True

    def _run_command(self, cmd_message: Any, cmd_data: Optional[bytes]) -> None:
        """Not used.

        Simply respect the interface.
        """

    def _receive_message(self, timeout_in_s: float) -> None:
        """Not used.

        Simply respect the interface.
        """



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
SocketCan to trc file logger
****************************

:module: socketcan_to_trc

:synopsis: Log SocketCAN messages into peak can trc file format.

.. currentmodule:: socketcan_to_trc

"""

import logging
import subprocess
import sys
import time
from textwrap import dedent

try:
    import can
except ImportError as e:
    raise ImportError(f"{e.name} dependency missing, consider installing pykiso with 'pip install pykiso[can]'")


log = logging.getLogger(__name__)


TRC_HEADER = """\
        ;$FILEVERSION=2.0
        ;$STARTTIME={starttime_days:12.3f}
        ;$COLUMNS=N,O,T,I,d,L,D
        ;
        ;   {filename:s}
        ;   start time: {starttime_str:s}
        ;   Generated by socketcan_to_trc(python)
        ;-------------------------------------------------------------------------------
        ;   Connection                                Bit rate
        ;   {can_name:s}                                      {ip_link_show:s}
        ;-------------------------------------------------------------------------------
        ; Glossary:
        ;   Direction of Message:
        ;     Rx: The frame was received
        ;     Tx: The frame was transmitted
        ;
        ;   Type of message:
        ;     DT: CAN or J1939 data frame
        ;     FD: CAN FD data frame
        ;     FB: CAN FD data frame with BRS bit set (Bit Rate Switch)
        ;     FE: CAN FD data frame with ESI bit set (Error State Indicator)
        ;     BI: CAN FD data frame with both BRS and ESI bits set
        ;     RR: Remote Request Frame
        ;     ST: Hardware Status change
        ;     ER: Error Frame
        ;     EV: Event. User-defined text, begins directly after 2-digit type indicator
        ;-------------------------------------------------------------------------------
        ;   Message   Time    Type ID     Rx/Tx
        ;   Number    Offset  |    [hex]  |  Data Length Code
        ;   |         [ms]    |    |      |  |  Data [hex]
        ;   |         |       |    |      |  |  |
        ;---+-- ------+------ +- --+----- +- +- +- -- -- -- -- -- -- --
        """


class SocketCan2Trc(can.Listener):
    """Creates a logfile containing CAN-BUS messages in the PEAK TRC format.
    Currently the only difference is the RX/TX column which is always RX.
    """

    def __init__(self, can_name: str, trc_file_name: str):
        """Initialise the logger

        :param can_name: socket ip link name
        :param trc_file_name: filename or "-" for stdout
        """
        self.started = False
        self.num = 0
        self.trc_file_name = trc_file_name
        self.can_name = can_name
        self.trc_file = sys.stdout
        self.starttime = self.get_start_time()

    def stop(self):
        """cleanup logger"""
        if not self.started:
            return
        # Remove the listener else the Can Notifier stop function will call this stop function
        # which then call again the Can Notifier again etc, and python is in an infinite loop
        self.can_notifier.remove_listener(self)
        self.can_notifier.stop()
        self.can_notifier = None
        if self.trc_file != sys.stdout:
            self.trc_file.close()
        self.bus = None
        self.started = False

    def start(self):
        """start logging"""
        if self.started:
            return

        self.bus = can.ThreadSafeBus(self.can_name, bustype="socketcan", fd=True)
        self.can_notifier = can.Notifier(self.bus, [self], timeout=1.0, loop=None)

        self.open_trc_file()

        self.starttime = self.get_start_time()
        ip_link_show = self.get_ip_link_show()
        starttime_days = self.get_start_time_days()

        self.log_trc_header(starttime_days, "Today", ip_link_show)
        self.started = True

    def on_message_received(self, msg: can.Message = None):
        """Increment message counter and log message.

        Override the can.Notifier.on_message_received.

        :param msg: message to log
        """
        self.log_can_frame(self.num, msg)
        self.num += 1

    @staticmethod
    def get_start_time():
        """return the system start time in float to have a offset of the time column

        :return: start time as an offset
        """
        return time.time()

    @staticmethod
    def get_start_time_days():
        """:return:  the days in double since 30.12.1899"""
        return time.time() / 60 / 60 / 24 + 25569

    def log_trc_header(self, starttime_days: float, starttime_str: str, ip_link_show: str):
        """Write the header of the TRC file

        :param starttime_days: number of days since 30.12.1899
        :param starttime_str: starttime string ( currently not used by parsers )
        :param ip_link_show: output of "ip -d link show" as an alternative to the pean information
        """
        print(
            dedent(TRC_HEADER).format(
                starttime_days=starttime_days,
                filename=self.trc_file_name,
                starttime_str=starttime_str,
                can_name=self.can_name,
                ip_link_show=ip_link_show,
            ),
            file=self.trc_file,
            flush=False,
        )

    def open_trc_file(self):
        """If trc_file_name is not stdout, open file trc-file"""

        # with not useful here
        # pylint: disable=R1732
        if self.trc_file_name != "-":
            self.trc_file = open(self.trc_file_name, mode="w", encoding="utf-8")

    @staticmethod
    def get_type(can_frame: can.Message):
        """return the TRC specific type of CAN-Frame

        :return: type string in TRC file
        """
        ret = ""
        if can_frame.is_remote_frame:
            ret = "RR"
        elif can_frame.is_error_frame:
            ret = "ER"
        elif not can_frame.is_fd:
            ret = "DT"
        elif can_frame.bitrate_switch and can_frame.error_state_indicator:
            ret = "BI"
        elif can_frame.bitrate_switch:
            ret = "FB"
        elif can_frame.error_state_indicator:
            ret = "FE"
        else:
            ret = "FD"

        return ret

    def log_can_frame(self, num: int, can_frame: can.Message):
        """Write a row to TRC file

        :param num: message number
        :param can_frame: can frame to log
        """
        if can_frame.is_error_frame:
            log.internal_warning("is errorframe")
        txt = "{num:7d} {time_msec:13.3f} {type:<6} {can_id:04x} RX {len:<2d}"
        print(
            txt.format(
                num=num,
                time_msec=(can_frame.timestamp - self.starttime) * 1000,
                type=self.get_type(can_frame),
                can_id=can_frame.arbitration_id,
                len=can_frame.dlc,
            ),
            file=self.trc_file,
            flush=False,
            end="",
        )

        if not can_frame.is_remote_frame:
            # f-string doesn't have formatting -> disable
            # pylint: disable=C0209
            for i in range(can_frame.dlc):
                print(
                    " {data:02X}".format(data=can_frame.data[i]),
                    file=self.trc_file,
                    end="",
                    flush=False,
                )

        print("", file=self.trc_file, flush=True)

    def get_ip_link_show(self):
        ''':return: the formatted output of "ip -d link show"'''
        complete_process = subprocess.run(
            ["ip", "-d", "link", "show", self.can_name],
            capture_output=True,
            check=True,
        )
        if complete_process.returncode != 0:
            raise RuntimeError("ip -d link show failed: " + complete_process.stdout.decode("ascii"))
        txt = complete_process.stdout.decode("ascii")
        lines = txt.splitlines(keepends=True)
        for i in range(1, len(lines), 1):
            lines[i] = ";                                         " + str(lines[i])
        return "".join(lines)



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
Create a Step report
********************

:module: assert_step_report

:synopsis: Provide a detailed test view containing:
    - test name
    - test description
    - date of execution + elapsed time
    - information gathered during test
    - assertion detail: data_in, variable, name of the data_in, expected, message

.. currentmodule:: assert_step_report

"""
import functools
import inspect
import linecache
import logging
import re
import sys
import traceback
import types
import typing
import unittest
from collections import OrderedDict
from dataclasses import dataclass, field
from datetime import datetime
from pathlib import Path
from typing import Dict, List, Union
from unittest.case import TestCase, _SubTest

import jinja2

from pykiso.test_coordinator.test_case import BasicTest

from .text_result import BannerTestResult
from .xml_result import TestInfo, XmlTestResult

log = logging.getLogger(__name__)


# Global variables
# Store the Result Step report
ALL_STEP_REPORT = OrderedDict()
# Step result keys used by Jinja for columns name
REPORT_KEYS = [
    "message",
    "var_name",
    "expected_result",
    "actual_result",
    "succeed",
    "failure_log",
    "description",
]
DEFAULT_TEST_METHOD = ["setup", "teardown", "handle_interaction"]
# Parent method being reported ; Ignore sub call (assert in an assert)
_FUNCTION_TO_APPLY = r"|".join(["test", *DEFAULT_TEST_METHOD])
#: content all assertion methods where the checked value is not shown
MUTE_CONTENT_ASSERTION = ["assertIsInstance", "assertNotIsInstance"]

# Jinja template location
SCRIPT_PATH = str(Path(__file__).resolve().parent)
REPORT_TEMPLATE = "templates/report_template.html.j2"
UNITTEST_DEPRECATED_FUNCTION_MAPPING: dict[str:str] = {
    "assertEqual": "assertEquals",
    "assertNotEqual": "assertNotEquals",
    "assertAlmostEqual": "assertAlmostEquals",
    "assertNotAlmostEqual": "assertNotAlmostEquals",
    "assertRaisesRegex": "assertRaisesRegexp",
    "assertRegex": "assertRegexpMatches",
    "assertNotRegex": "assertNotRegexpMatches",
}


@dataclass
class StepReportData:
    # Store additional data fetched during test for
    # the step-report (even if not activated)
    header: dict = field(default_factory=dict)
    # Message for step in report
    message: str = ""
    # Test succeed flag
    # Set to false if one or more steps fail
    success: bool = True
    # Error message on step fail
    last_error_message: str = ""
    # Current report table
    current_table: typing.Optional[str] = None


def _get_variable_name(f_back: types.FrameType, assert_name: str) -> str:
    """Get the input parameter name to the assert method.

    According to unittest.TestCase, it is always placed first.
    That function read the source code as a text to find it.

    :param f_back: frame calling the assert method
    :param: assert_name: current assert method name

    return: variable name
    """
    expected_varname = ""
    # Get source code lines
    file = inspect.getsourcefile(f_back)

    # Get current line number
    line_no = f_back.f_lineno

    # For multiline statements, the current frame points to the last line in python < 3.8
    # and to the first line in python >= 3.8
    first_line = sys.version_info >= (3, 8)

    # Get line content
    line = linecache.getline(file, line_no)
    lines = line

    if first_line:  # pragma: no cover
        # Read top down, count parentheses to find last line of assert statement
        open_parentheses, closed_parentheses = line.count("("), line.count(")")
        while open_parentheses != closed_parentheses:
            line_no += 1
            line = linecache.getline(file, line_no)
            lines += line
            open_parentheses += line.count("(")
            closed_parentheses += line.count(")")
    else:
        # Read bottom up until the assert method is reached
        while assert_name not in line:
            line_no -= 1
            previous_line = linecache.getline(file, line_no)
            lines = previous_line + line
            line = previous_line

    # remove line breaks and spaces
    lines = lines.replace("\n", "").replace(" ", "")

    # remove current assert function name
    lines = re.split(rf".*{assert_name}\(", lines)[1]

    # Check the variable start names
    if re.findall(r"^[a-zA-Z]", lines):
        # Get variable name
        expected_varname = re.findall(r"[a-zA-Z0-9_]+", lines)[0]

    return expected_varname


def _get_expected(func_name: str, arguments: dict) -> str:
    """Get the assertion purpose and the expected value

    :param func_name: name of the assertion function
    :param arguments: argument of the function
        eg: {arg_name: value}

    :return: expected value
    """
    # Get assertion purpose (eg: get Equal from assertEqual)
    name = re.findall(r"([A-Z][a-z]+)", func_name)
    expected = " ".join(name)

    # Get expected value and parameters if exist
    # eg: assertAlmostEqual(50, 60, msg="message", delta=10)
    # but assertTrue(var1, "message") -> nothing else to do
    if len(arguments) >= 3:
        # Parameters not known in advance.
        # Remove the "input value" and the "msg".
        arguments_keys = list(arguments.keys())
        args = arguments.copy()
        args.pop("msg", None)
        # "input value" always 1st position
        args.pop(arguments_keys[0], None)
        # "expected value" always 2nd position
        expected_val = args.pop(arguments_keys[1], None)
        # Add expected value to string
        expected += f" to {str(expected_val)}"
        # Add the additional parameter if exist
        if args:
            # only param are left in args
            expected += "; with"
            for key, value in args.items():
                expected += f" {key}={value},"
            # remove the coma from last loop
            expected = expected[:-1]

    return expected


def _prepare_report(test: unittest.case.TestCase, test_name: str) -> None:
    """Make ALL_STEP_REPORT variable ready for update

    Create the tree if required, otherwise, does nothing
    - CurrentTestClass
        - header: additional data
        - description: test description from test_run
        - file_path: test file location
        - time_result: start/end/elapsed time
        - test_list: store the steps result
            - test_name: list of steps

    :param test: test to be reported
    :param test_name: name of the function being tested
    """
    global ALL_STEP_REPORT

    test_class_name = type(test).__name__

    # Create the testClass
    if not ALL_STEP_REPORT.get(test_class_name):
        ALL_STEP_REPORT[test_class_name] = OrderedDict()
        # Add test succeed flag
        ALL_STEP_REPORT[test_class_name]["succeed"] = True
        # Add header (mutable object -> dictionary fed during test)
        ALL_STEP_REPORT[test_class_name]["header"] = test.step_report.header
        # Add description of the test -> Always test_run
        ALL_STEP_REPORT[test_class_name]["description"] = test.__doc__ or "Not provided"
        # Add test file path
        ALL_STEP_REPORT[test_class_name]["file_path"] = inspect.getfile(type(test))
        # Store the result (start, stop, elapsed time)
        ALL_STEP_REPORT[test_class_name]["time_result"] = OrderedDict()
        ALL_STEP_REPORT[test_class_name]["time_result"]["Start Time"] = 0
        # Store the tests list
        ALL_STEP_REPORT[test_class_name]["test_list"] = OrderedDict()

    # Create the current test step storage
    if not ALL_STEP_REPORT[test_class_name]["test_list"].get(test_name):
        test_method = getattr(test, test_name, None)
        test_description = (test_method.__doc__ or test._testMethodDoc or "").split(" [")[0]
        ALL_STEP_REPORT[test_class_name]["test_list"][test_name]: Dict[
            str,
            Union[str, List[List[Dict[str, Union[str, bool]]]], List[List[str]]],
        ] = {
            "description": test_description,
            "steps": [[]],
            "unexpected_errors": [[]],
        }


def _add_step(
    test_class_name: str,
    test_name: str,
    message: str,
    var_name: str,
    expected: typing.Any,
    received: typing.Any,
    failure_log: str,
):
    global ALL_STEP_REPORT, REPORT_KEYS

    ALL_STEP_REPORT[test_class_name]["test_list"][test_name]["steps"][-1].append(
        dict(
            zip(
                REPORT_KEYS,
                [
                    message,
                    var_name,
                    expected,
                    received,
                    True,
                    failure_log,
                ],
            )
        )
    )


def determine_parent_test_function(test_name: str) -> str:
    """Determine the parent test function.

    This function attached the nested assertion to the correct parent test
    function.

    :param test_name: current test function

    :return: parent test function
    """
    # the function respect the default test method pattern or in default test function
    if test_name.lower() in DEFAULT_TEST_METHOD or test_name.startswith("test_"):
        return test_name

    # collect all methods called from the stack to get the parent test function
    methods = [frame.function for frame in inspect.stack()]

    # determine if the parent test function is a fixture or a test function
    fixture = [method for method in methods if method.lower() in DEFAULT_TEST_METHOD]
    test_function = [method for method in methods if method.startswith("test_")]

    if fixture:
        return fixture.pop()

    if test_function:
        return test_function.pop()


def assert_decorator(assert_method: types.MethodType):
    """Decorator to gather assertion information

    - MyTestClass
        - header: additional data
        - description: test purpose
        - file_path: test location
        - time_result: start/end/elapsed time
        - test_list: store the steps result
            - setUp : list of assert
            - test_run: list of assert

    .. note:: header is based on the variable ``step_report_header`` and
        description is based on the docstring of the test_run method

    .. code:: python

        MyTest(pykiso.BasicTest):
            def test_run(self):
                '''Here is my test description'''
                self.step_report.header["Voltage"] = 5

    :param func: function to decorate (expected assert method)

    return: The func output if it exists. Otherwise, None
    """

    @functools.wraps(assert_method)
    def func_wrapper(*args, **kwargs):
        """Decorator Main
        Fetch the assert step: message, var_name, expected, received

        return: The assertion method output if it exists. Otherwise, None
        """
        global _FUNCTION_TO_APPLY
        original_logic = False
        try:
            # Context
            currentframe = inspect.currentframe()
            f_back = currentframe.f_back

            # Case where the assert is called in unittest so we don't want to add any logic
            if f_back.f_globals["__name__"] == "unittest.case":
                original_logic = True
                return assert_method(*args, **kwargs)

            test_name = determine_parent_test_function(f_back.f_code.co_name)
            test_case_inst: TestCase = assert_method.__self__
            test_class_name = type(test_case_inst).__name__
            assert_name = assert_method.__name__

            # filter parent call, only known function recorded
            parent_method = re.findall(_FUNCTION_TO_APPLY, test_name.lower())
            # get the decorated test fixture name (setUp, tearDown, ...)
            if parent_method and parent_method[0] == "handle_interaction":
                test_name = f_back.f_locals["func"].__name__

            # Assign variables to signature
            signature = inspect.signature(assert_method)
            arguments = signature.bind(*args, **kwargs).arguments
            test_name = test_case_inst.step_report.current_table or test_name
            # 1. Gather message, var_name, expected, received, failure_log
            # 1.1 Get message. default value: ""
            if test_case_inst.step_report.message:
                message = test_case_inst.step_report.message
                test_case_inst.step_report.message = ""
            else:
                message = arguments.get("msg", "")

            # ensure message is always present in the arguments
            # dictionary. (used in _get_expected)
            if not message and "msg" in signature.parameters:
                arguments["msg"] = ""

            # 1.2. Get 'received" value (Always 1st argument)
            assert_value = list(arguments.values())[0]
            received = assert_value if assert_name not in MUTE_CONTENT_ASSERTION else ""

            # 1.3. Get variable name
            found = False
            deprecated_try = False
            while not found:
                try:
                    var_name = _get_variable_name(f_back, assert_name)
                    found = True
                except IndexError:
                    f_back = f_back.f_back
                except TypeError:
                    # replace assert name with depecrated function assert name
                    if not deprecated_try and assert_name in UNITTEST_DEPRECATED_FUNCTION_MAPPING.keys():
                        deprecated_try = True
                        f_back = currentframe.f_back
                        assert_name = UNITTEST_DEPRECATED_FUNCTION_MAPPING[assert_name]
                        continue
                    log.error(f"Step report error: Variable name couldn't be find for {currentframe.f_back}")
                    var_name = "Variable name not found"
                    break

            # 1.4. Get Expected value
            expected = _get_expected(assert_name, arguments)

            # 1.5. Get the failure logs
            failure_log = traceback.format_exc() if test_case_inst.failureException else "None"

            # 2. Update report data
            # 2.1 Ensure report ready for update
            _prepare_report(test_case_inst, test_name)

            # 2.2. Add new step
            _add_step(
                test_class_name,
                test_name,
                message,
                var_name,
                expected,
                received,
                failure_log,
            )

        except Exception as e:
            if original_logic:
                raise e
            log.exception(f"Unable to update Step due to exception: {e}")

        # Run the assert method and mark the test as failed if the AssertionError is raised
        try:
            return assert_method(*args, **kwargs)
        except test_case_inst.failureException as e:
            log.error(f"Assert step exception: {e}")
            test_case_inst.step_report.last_error_message = f"{e}"
            if parent_method:
                ALL_STEP_REPORT[test_class_name]["test_list"][test_name]["steps"][-1][-1]["succeed"] = False
                ALL_STEP_REPORT[test_class_name]["succeed"] = False
                ALL_STEP_REPORT[test_class_name]["test_list"][test_name]["steps"][-1][-1][
                    "failure_log"
                ] = traceback.format_exc()

            test_case_inst.step_report.success = False

            # repeat the assertion failure as first element in the traceback
            frame = currentframe.f_back
            tb = types.TracebackType(None, frame, frame.f_lasti, frame.f_lineno)
            raise e.with_traceback(tb)

    return func_wrapper


def _parse_timestamp(timestamp: float) -> str:
    """Parse timestamp to date: alike 31/12/2021 23:59:59

    :param timestamp: Seconds since 01/01/1970

    :return: the date string
    """
    dt = datetime.fromtimestamp(timestamp)
    return dt.strftime("%d/%m/%y %H:%M:%S")


def is_test_success(test: dict) -> bool:
    """Check if test was successful.

    :param tests: test
    :return: True if each step in a test was successful and no unexpected error
        was raised else False
    """
    return all([step["succeed"] for step in test["steps"][-1]]) and not test.get("unexpected_errors")[-1]


jinja_template_functions = {
    "is_test_success": is_test_success,
}


def generate_step_report(
    test_result: Union[BannerTestResult, XmlTestResult],
    output_file: str,
) -> None:
    """Generate the HTML step report based on Jinja2 template

    :param test_result: Result of tests to generate the report from
    :param output_file: Report output file path
    """
    global ALL_STEP_REPORT, SCRIPT_PATH, REPORT_TEMPLATE
    test_result.stream.writeln("Generating HTML reports...")

    succeeded_tests = test_result.successes + test_result.expectedFailures
    failed_test = test_result.failures + test_result.errors + test_result.unexpectedSuccesses
    # Update info for each test
    for test_case in succeeded_tests + failed_test:
        if isinstance(test_case, (TestCase, TestInfo)):
            # Case of success
            test_info = test_case
        elif isinstance(test_case, tuple) and isinstance(test_case[0], (TestCase, TestInfo)):
            # Case of non success
            test_info = test_case[0]

        if isinstance(test_info, _SubTest):
            class_name = test_info.test_case.__class__.__name__
            start_time = _parse_timestamp(test_info.test_case.start_time)
            stop_time = _parse_timestamp(test_info.test_case.stop_time)
            elapsed_time = test_info.test_case.elapsed_time
            test_method_name = test_info.test_case._testMethodName
        elif isinstance(test_info, TestCase):
            class_name = test_info.__class__.__name__
            start_time = _parse_timestamp(test_info.start_time)
            stop_time = _parse_timestamp(test_info.stop_time)
            elapsed_time = test_info.elapsed_time
            test_method_name = test_info._testMethodName
        elif isinstance(test_info, TestInfo):
            # test_info is TestInfo
            class_name = test_info.test_name.split(".")[-1]
            start_time = _parse_timestamp(test_info.test_result.start_time)
            stop_time = _parse_timestamp(test_info.test_result.stop_time)
            elapsed_time = test_info.elapsed_time
            test_method_name = test_info.test_id.split(".")[-1]
        # Update test_case
        if class_name in ALL_STEP_REPORT:
            ALL_STEP_REPORT[class_name]["time_result"]["Start Time"] = start_time
            ALL_STEP_REPORT[class_name]["time_result"]["End Time"] = stop_time
            ALL_STEP_REPORT[class_name]["time_result"]["Elapsed Time"] = round(elapsed_time, 2)
            if test_case in test_result.errors:
                for test_name in ALL_STEP_REPORT[class_name]["test_list"].keys():
                    name = test_name if test_name != test_method_name else test_method_name
                    test_list = ALL_STEP_REPORT[class_name]["test_list"][name]
                    test_list["unexpected_errors"][-1].append(test_case[1])
                    ALL_STEP_REPORT[class_name]["succeed"] = False
    # Render the source template
    render_environment = jinja2.Environment(loader=jinja2.FileSystemLoader(SCRIPT_PATH), autoescape=True)
    template = render_environment.get_template(REPORT_TEMPLATE)
    template.globals.update(jinja_template_functions)
    output_text = template.render({"ALL_STEP_REPORT": ALL_STEP_REPORT})

    output_file = Path(output_file).resolve()
    output_file.parent.mkdir(parents=True, exist_ok=True)
    # Write the output into the output file
    with output_file.open("w") as report_file:
        report_file.write(output_text)


def add_retry_information(
    test: BasicTest,
    result_test: bool,
    retry_nb: int,
    max_try: int,
    exc: Exception,
) -> None:
    """Add information in the step report if a test fails and is retried.

    :param test: test failed
    :param result_test: result of the tests of the class before the retry
    :param retry_nb: number of the current try
    :param max_try: maximum tries that will be done
    :param exc: exception caught
    """
    global ALL_STEP_REPORT
    test_class_name = type(test).__name__
    test_information = ALL_STEP_REPORT[test_class_name]["test_list"][test._testMethodName]
    # Add information about the number of try
    test_information["number_try"] = retry_nb + 1
    test_information["max_try"] = max_try
    # Add another list to steps to differentiate the try
    test_information["steps"].append([])
    # We add the error raised if it was not an assertion error
    if not isinstance(exc, AssertionError):
        test_information["unexpected_errors"][-1].append(traceback.format_exc())
    test_information["unexpected_errors"].append([])
    # Go back to the state before executing the test
    ALL_STEP_REPORT[test_class_name]["succeed"] = result_test



##########################################################################
# Copyright (c) 2010-2025 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
Profiling
*********

:module: profiling

:synopsis: add helper functions for profiling

.. currentmodule:: profiling

"""

import json
from contextlib import contextmanager
from pathlib import Path

from viztracer import VizTracer
from viztracer.vcompressor import VCompressor

tracer = VizTracer()


def get_tracer() -> VizTracer:
    """
    Get the VizTracer instance

    :return: the VizTracer instance
    """
    return tracer


@contextmanager
def profile_manager(filename: str = "result.json", compress: bool = False):
    """
    Context manager for VizTracer

    :param filename: the name of the file to save the result (default: result.json)
    :param compress: compress the result file if True and save it with .cvf extension (default: False)
    """
    tracer.start()
    try:
        yield None
    finally:
        tracer.stop()
        tracer.save(filename)
        if compress:
            compress_trace(filename)


def compress_trace(source_file: str):
    """
    Compress the trace file and delete the original file

    :param source_file: the name of the file to compress
    """
    compressor = VCompressor()
    input_filename = Path(source_file)
    output_filename = input_filename.with_suffix(".cvf").as_posix()
    with open(source_file) as f:
        data = json.load(f)
        compressor.compress(data, output_filename)
    input_filename.unlink()  # delete the original file


def profile(filename: str = "", compress: bool = False):
    """
    Decorator for VizTracer

    :param filename: the name of the file to save the result (default: module_name_function_name.json)
    :param compress: compress the result file if True and save it with .cvf extension (default: False)
    """

    def decorator(func: callable):
        result_filename = func.__module__ + "_" + func.__name__ + ".json" if filename == "" else filename

        def wrapper(*args, **kwargs):
            tracer.start()
            result = func(*args, **kwargs)
            tracer.stop()
            if result_filename:
                tracer.save(result_filename)
            if compress:
                compress_trace(result_filename)
            return result

        return wrapper

    return decorator



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
Generic Test
************

:module: test_case

:synopsis: Basic extensible implementation of a TestCase, and of a Remote
    TestCase for Message Protocol / TestApp usage.

.. currentmodule:: test_case

.. note::
    TODO later on will inherit from a metaclass to get the id parameters

"""
from __future__ import annotations

import functools
import logging
import unittest
import warnings
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Type, Union

import pykiso.test_result.assert_step_report as step_report

from .. import message
from ..auxiliary import AuxiliaryInterface
from .test_message_handler import test_app_interaction

if TYPE_CHECKING:
    from .test_suite import BaseTestSuite

log = logging.getLogger(__name__)


class BasicTest(unittest.TestCase):
    """Base for test-cases."""

    def __init__(
        self,
        test_suite_id: int,
        test_case_id: int,
        aux_list: Union[List[AuxiliaryInterface], None],
        setup_timeout: Union[int, None],
        run_timeout: Union[int, None],
        teardown_timeout: Union[int, None],
        test_ids: Union[dict, None],
        tag: Union[Dict[str, List[str]], None],
        *args: Any,
        **kwargs: Any,
    ):
        """Initialize generic test-case.

        :param test_suite_id: test suite identification number
        :param test_case_id: test case identification number
        :param aux_list: list of used auxiliaries
        :param setup_timeout: maximum time (in seconds) used to wait
            for a report during setup execution
        :param run_timeout: maximum time (in seconds) used to wait for
            a report during test_run execution
        :param teardown_timeout: the maximum time (in seconds) used to
            wait for a report during teardown execution
        :param test_ids: jama references to get the coverage
            eg: {"Component1": ["Req1", "Req2"], "Component2": ["Req3"]}
        :param tag: dictionary allowing users to filter the tests based
            on the keys and their value.
        """
        # Initialize base class
        super().__init__(*args, **kwargs)
        # Save list of test auxiliaries to use (already initialize)
        self.test_auxiliary_list = aux_list or []
        # Save test information
        self.test_suite_id = test_suite_id
        self.test_case_id = test_case_id
        self.test_ids = test_ids
        self.tag = tag
        self.start_time = self.stop_time = self.elapsed_time = 0
        if any([setup_timeout, run_timeout, teardown_timeout]) and not isinstance(self, RemoteTest):
            log.warning("BasicTest does not support test timeouts, it will be discarded")

        self._properties = None

    def __str__(self):
        return "%s (%s)" % (self._testMethodName, self.__class__.__name__)

    def cleanup_and_skip(self, aux: AuxiliaryInterface, info_to_print: str) -> None:
        """Cleanup auxiliary and log reasons.

        :param aux: corresponding auxiliary to abort
        :param info_to_print: A message you want to print while cleaning up the test
        """
        # Log error message
        log.critical(info_to_print)

        # Send aborts to corresponding auxiliary
        if aux.send_abort_command(timeout=10) is not True:
            log.critical(f"Error occurred during abort command on auxiliary {aux}")

        self.fail(info_to_print)

    def setUp(self) -> None:
        """Startup hook method to execute code before each test method."""
        pass

    def tearDown(self) -> None:
        """Closure hook method to execute code after each test method."""
        pass

    @property
    def properties(self):
        return self._properties

    @properties.setter
    def properties(self, value):
        if isinstance(self._properties, dict) and isinstance(value, dict):
            self._properties.update(value)
        else:
            self._properties = value

    @properties.deleter
    def properties(self):
        del self._properties

    # Overwrite the deprecate function to keep the signature for step report
    @staticmethod
    def _deprecate(original_func):
        @functools.wraps(original_func)
        def deprecated_func(*args, **kwargs):
            warnings.warn("Please use {0} instead.".format(original_func.__name__), DeprecationWarning, 2)
            return original_func(*args, **kwargs)

        return deprecated_func

    failUnlessEqual = assertEquals = _deprecate(unittest.TestCase.assertEqual)
    failIfEqual = assertNotEquals = _deprecate(unittest.TestCase.assertNotEqual)
    failUnlessAlmostEqual = assertAlmostEquals = _deprecate(unittest.TestCase.assertAlmostEqual)
    failIfAlmostEqual = assertNotAlmostEquals = _deprecate(unittest.TestCase.assertNotAlmostEqual)
    failUnless = assert_ = _deprecate(unittest.TestCase.assertTrue)
    failUnlessRaises = _deprecate(unittest.TestCase.assertRaises)
    failIf = _deprecate(unittest.TestCase.assertFalse)
    assertRaisesRegexp = _deprecate(unittest.TestCase.assertRaisesRegex)
    assertRegexpMatches = _deprecate(unittest.TestCase.assertRegex)
    assertNotRegexpMatches = _deprecate(unittest.TestCase.assertNotRegex)


class RemoteTest(BasicTest):
    """Base test-cases for Message Protocol / TestApp usage."""

    response_timeout: int = 10

    def __init__(
        self,
        test_suite_id: int,
        test_case_id: int,
        aux_list: Union[List[AuxiliaryInterface], None],
        setup_timeout: Union[int, None],
        run_timeout: Union[int, None],
        teardown_timeout: Union[int, None],
        test_ids: Union[dict, None],
        tag: Union[Dict[str, List[str]], None],
        *args: Any,
        **kwargs: Any,
    ):
        """Initialize TestApp test-case.

        :param test_suite_id: test suite identification number
        :param test_case_id: test case identification number
        :param aux_list: list of used auxiliaries
        :param setup_timeout: maximum time (in seconds) used to wait
            for a report during setup execution
        :param run_timeout: maximum time (in seconds) used to wait for
            a report during test_run execution
        :param teardown_timeout: the maximum time (in seconds) used to
            wait for a report during teardown execution
        :param test_ids: jama references to get the coverage
            eg: {"Component1": ["Req1", "Req2"], "Component2": ["Req3"]}
        :param tag: dictionary containing lists of variants and/or test
            levels when only a subset of tests needs to be executed
        """
        super().__init__(
            test_suite_id,
            test_case_id,
            aux_list,
            setup_timeout,
            run_timeout,
            teardown_timeout,
            test_ids,
            tag,
            *args,
            **kwargs,
        )
        self.setup_timeout = setup_timeout or RemoteTest.response_timeout
        self.run_timeout = run_timeout or RemoteTest.response_timeout
        self.teardown_timeout = teardown_timeout or RemoteTest.response_timeout

    @test_app_interaction(message_type=message.MessageCommandType.TEST_CASE_SETUP, timeout_cmd=5)
    def setUp(self) -> None:
        """Startup hook method to execute code before each test method."""
        pass

    @test_app_interaction(message_type=message.MessageCommandType.TEST_CASE_RUN, timeout_cmd=5)
    def test_run(self) -> None:
        """Hook method from unittest in order to execute test case."""
        pass

    @test_app_interaction(
        message_type=message.MessageCommandType.TEST_CASE_TEARDOWN,
        timeout_cmd=5,
    )
    def tearDown(self) -> None:
        """Closure hook method to execute code after each test method."""
        pass


def define_test_parameters(
    suite_id: int = 0,
    case_id: int = 0,
    aux_list: List[AuxiliaryInterface] = None,
    setup_timeout: Optional[int] = None,
    run_timeout: Optional[int] = None,
    teardown_timeout: Optional[int] = None,
    test_ids: Optional[dict] = None,
    tag: Optional[Dict[str, List[str]]] = None,
):
    """Decorator to fill out test parameters of the BasicTest and RemoteTest automatically."""

    def generate_modified_class(
        DecoratedClass: Type[Union[BasicTest, BaseTestSuite]],
    ) -> Type[Union[BasicTest, BaseTestSuite]]:
        """For basic test-case, generates the same class but with the test IDs
        already filled. It works as a partially filled-out call to the __init__ method.
        """

        class NewClass(DecoratedClass):
            """Modified {DecoratedClass.__name__}, with the __init__ method
            already filled out with the following test-parameters:
            Suite ID:    {suite_id}
            Case ID:     {case_id}
            Auxiliaries: {auxes}
            setup_timeout: {setup_timeout}
            run_timeout: {run_timeout}
            teardown_timeout: {teardown_timeout}
            test_ids: {test_ids}
            tag: {tag}
            """

            @functools.wraps(DecoratedClass.__init__)
            def __init__(self, *args, **kwargs):
                super(NewClass, self).__init__(
                    suite_id,
                    case_id,
                    aux_list,
                    setup_timeout,
                    run_timeout,
                    teardown_timeout,
                    test_ids,
                    tag,
                    *args,
                    **kwargs,
                )

        NewClass.__doc__ = DecoratedClass.__doc__

        # Used to display the current test module in the test result
        NewClass.__module__ = DecoratedClass.__module__
        # Passing the name of the decorated class to the new returned class
        # in order to get the test case name and references, i.e. suite_id and case_id
        # in the test results in the console and in the report.
        # Changing __name__ is necessary to make the test name appear in the test results in the console.
        # Changing __qualname__ is necessary to make the test name appear in the test results in the report.
        ids = "" if suite_id == 0 and case_id == 0 else f"-{suite_id}-{case_id}"
        NewClass.__name__ = f"{DecoratedClass.__name__}{ids}"
        NewClass.__qualname__ = f"{DecoratedClass.__qualname__}{ids}"
        return NewClass

    return generate_modified_class


def retry_test_case(
    max_try: int = 2,
    rerun_setup: bool = False,
    rerun_teardown: bool = False,
    stability_test: bool = False,
):
    """Decorator: retry mechanism for testCase.

    The aim is to cover the 2 following cases:

        - Unstable test : get the test pass within the {max_try} attempt

        - Stability test : run {max_try} time the test expecting no error

    The **retry_test_case** comes with the possibility to re-run the setUp and
    tearDown methods automatically.

    :param max_try: maximum number of try to get the test pass.
    :param rerun_setup: call the "setUp" method of the test.
    :param rerun_teardown: call the "tearDown" method of the test.
    :param stability_test: run {max_try} time the test and raise an exception if an error occurs.

    :return: None, a testCase is not supposed to return anything.
    :raise Exception: if stability_test, the exception that occurred during the execution; if
        not stability_test, the exception that occurred at the last try.
    """

    def decorator(func):
        @functools.wraps(func)
        def func_wrapper(self: BasicTest) -> None:
            # track the current execution for logging
            current_execution = None
            test_class_name = type(self).__name__
            # Prepare report for the current test so we can get the current result for the class
            if getattr(self, "step_report", False) and self.step_report.header:
                step_report._prepare_report(self, self._testMethodName)
                result_test = step_report.ALL_STEP_REPORT[test_class_name]["succeed"]

            for retry_nb in range(1, max_try + 1):
                try:
                    # by the 2nd attempt, end the test with the teardown and start with setUp
                    if retry_nb > 1:
                        if rerun_teardown:
                            current_execution = self.tearDown
                            self.tearDown()
                        if rerun_setup:
                            current_execution = self.setUp
                            self.setUp()

                    # run the method (eg: test_run(self))
                    current_execution = func
                    func(self)
                    if not stability_test:
                        break
                    else:
                        # Clearly separate tests
                        log.info(f">>>>>>>>>> Stability test {retry_nb}/{max_try} succeed <<<<<<<<<<")

                except Exception as e:
                    # log: test_name (class), method (setUp, test_run, tearDown) and the error.
                    log.warning(f"{self.__class__.__name__}.{current_execution.__name__} failed with exception: {e}.")

                    # raise the exception that occurred during the latest attempt
                    if retry_nb == max_try or stability_test:
                        log.error(f">>>>>>>>>> Test {retry_nb}/{max_try} failed <<<<<<<<<<")
                        raise e
                    elif getattr(self, "step_report", False) and self.step_report.header:
                        step_report.add_retry_information(self, result_test, retry_nb, max_try, e)

                    # print counter only after failing test to avoid spamming the console
                    log.info(f">>>>>>>>>> Attempt: {retry_nb + 1}/{max_try} <<<<<<<<<<")

        return func_wrapper

    return decorator


def xray(test_key: str, req_id: str | None = None) -> None:
    """Decorator: to mark the test to import the JUnit xml results into xray

    :param test_key: the xray ticket id linked to be linked to the test
    :param req_id: the requirement ticket id to be linked to the test
    """

    def decorator(func):
        @functools.wraps(func)
        def func_wrapper(*args, **kwargs):
            args[0].properties = {
                "test_key": test_key,
                "req_id": req_id,
                "test_description": func.__doc__,
            }
            result = func(*args, **kwargs)
            return result

        return func_wrapper

    return decorator



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################


"""
TestRail API
************

:module: api

:synopsis: implement all the APIs used to interact with TestRail REST
    API

.. currentmodule:: api

"""

import base64
import concurrent.futures
import json
from enum import Enum
from typing import Callable, List

import requests


class ContentType(Enum):
    """Store all available header's content-type."""

    #: JSON content type
    JSON: str = "application/json"


class ApiVersion(Enum):
    """Store all available TestRail rest API version."""

    #: TestRail's V2 API
    V2: str = "index.php?/api/v2/"


class HttpError(Exception):
    """Raise if the status code from a HTTP request is different than
    200.
    """

    def __init__(self, status_code: str, reason: str) -> None:
        """Initialize error message attributes.

        :param status_code: response status code
        :param reason: response content
        """
        self.status_code = status_code
        self.reason = reason

    def __str__(self) -> str:
        """Exception message representation."""
        return f"HTTP request exit with status {self.status_code} : {self.reason}"


def handle_http_error(func: Callable) -> Callable:
    """Decorator used to handle HTTP request negative response

    :param func: inner handle function

    :return: the function returned value

    :raises Exception: if the response status code is different than
            200
    """

    def handle(*args: tuple, **kwargs: dict) -> requests.Response:
        """Handle sttaus code different than 200.

        :param args: positonnal arguments

        :return: kwargs named arguments

        :raises Exception: if the response status code is different than
                200
        """
        response = func(*args, **kwargs)

        if response.status_code != 200:
            raise HttpError(response.status_code, response.content)

        return response

    return handle


class HttpRequest:
    """Encapsulate all available HTTP requests."""

    @staticmethod
    @handle_http_error
    def get(url: str, headers: dict) -> requests.Response:
        """Send a GET HTTP request and return the received response.

        :param url: request target url
        :param headers: request's header (authorization, content-type)

        :return: received response from server

        :raises HttpError: if the response status code is different than
            200
        """
        return requests.get(url, headers=headers, verify=True)

    @staticmethod
    @handle_http_error
    def post(url: str, headers: dict, data: dict) -> requests.Response:
        """Send a POST HTTP request and return the received response.

        :param url: request target url
        :param headers: request's header
        :param data: request body

        :return: received response from server

        :raises HttpError: if the response status code is different than 200
        """
        return requests.post(url, headers=headers, data=data, verify=True)


class TestRailApi:
    """TestRail command API."""

    #: store the current API version in use
    API_VERSION: str = ApiVersion.V2.value

    @staticmethod
    def _construct_url(base_url: str, api_version: str, uri: str) -> str:
        """Construct the TestRail url.

        :param base_url: TestRail base url
        :param api_version: current api version in use
        :param uri: api command reference

        :return: request's url
        """
        return f"{base_url}/{api_version}{uri}" if not base_url.endswith("/") else f"{base_url}{api_version}{uri}"

    @staticmethod
    def _construct_header(user: str, password: str, content_type: str) -> dict:
        """Construct the request header by adding content and authorization
        part.

        :param user: user id
        :param password: user's password or API key
        :param content_type: reauest's content type (json,...)

        :return: request's fulfilled header
        """

        auth = str(
            base64.b64encode(bytes("%s:%s" % (user, password), "utf-8")),
            "ascii",
        ).strip()

        headers = {"Authorization": "Basic " + auth}
        headers["Content-Type"] = content_type
        return headers

    @classmethod
    def get_projects(cls, base_url: str, user: str, password: str) -> dict:
        """Retrieve all available projects under TestRail

        :param base_url: TestRail's base url
        :param user: user's session id
        :param password: user's password or API key

        :return: all available projects information
        """
        url = cls._construct_url(base_url, cls.API_VERSION, uri="get_projects")
        headers = cls._construct_header(user, password, ContentType.JSON.value)
        query_result = HttpRequest.get(url=url, headers=headers)
        return json.loads(query_result.content)

    @classmethod
    def get_suites(cls, base_url: str, user: str, password: str, project_id: int) -> dict:
        """Retrieve all available suite under the given project.

        :param base_url: TestRail's base url
        :param user: user's session id
        :param password: user's password
        :param project_id: project id in which to search for suites

        :return: all available suites information
        """
        url = cls._construct_url(base_url, cls.API_VERSION, uri=f"get_suites/{project_id}")
        headers = cls._construct_header(user, password, ContentType.JSON.value)
        query_result = HttpRequest.get(url=url, headers=headers)
        return json.loads(query_result.content)

    @classmethod
    def get_cases(
        cls,
        base_url: str,
        user: str,
        password: str,
        project_id: int,
        suite_id: int,
    ) -> dict:
        """Retrieve all available test cases under the given project and
        suite.

        :param base_url: TestRail's base url
        :param user: user's session id
        :param password: user's password
        :param project_id: targeted project id
        :param suite_id: targeted suite id

        :return: all available runs information
        """
        url = cls._construct_url(
            base_url,
            cls.API_VERSION,
            uri=f"get_cases/{project_id}&suite_id={suite_id}",
        )
        headers = cls._construct_header(user, password, ContentType.JSON.value)
        query_result = HttpRequest.get(url=url, headers=headers)
        return json.loads(query_result.content)

    @classmethod
    def async_get_cases(
        cls,
        base_url: str,
        user: str,
        password: str,
        project_id: int,
        suite_ids: List[int],
    ) -> List[dict]:
        """Asynchronously get test cases information.

        :param base_url: TestRail's base url
        :param user: user's session id
        :param password: user's password
        :param project_id: TestRail's project id
        :param suite_ids: all suite ids to search for test cases

        :return: all test case information
        """
        with concurrent.futures.ThreadPoolExecutor(max_workers=len(suite_ids)) as executor:
            future_pools = [
                executor.submit(
                    TestRailApi.get_cases,
                    base_url,
                    user,
                    password,
                    project_id,
                    suite_id,
                )
                for suite_id in suite_ids
            ]
            return [pool.result() for pool in concurrent.futures.as_completed(future_pools, timeout=20)]

    @classmethod
    def async_add_result_for_case(
        cls,
        base_url: str,
        user: str,
        password: str,
        run_id: int,
        case_ids: List[int],
        data: List[dict],
    ) -> List[dict]:
        """Asynchronously add results to all the given test cases.

        :param base_url: TestRail's base url
        :param user: user's session id
        :param password: user's password
        :param run_id: TestRail's run id
        :param case_id: all TestRail's case ids to update
        :param data: all case's results

        :return: all new test results
        """
        with concurrent.futures.ThreadPoolExecutor(max_workers=len(case_ids)) as executor:
            future_pools = [
                executor.submit(
                    cls.add_result_for_case,
                    base_url,
                    user,
                    password,
                    run_id,
                    case_id,
                    _data,
                )
                for case_id, _data in zip(case_ids, data)
            ]
        return [pool.result() for pool in concurrent.futures.as_completed(future_pools, timeout=20)]

    @classmethod
    def get_runs(cls, base_url: str, user: str, password: str, project_id: int) -> dict:
        """Retrieve all available runs under the given project.

        :param base_url: TestRail's base url
        :param user: user's session id
        :param password: user's password
        :param project_id: target project id

        :return: all available runs information
        """
        url = cls._construct_url(base_url, cls.API_VERSION, uri=f"get_runs/{project_id}")
        headers = cls._construct_header(user, password, ContentType.JSON.value)
        query_result = HttpRequest.get(url=url, headers=headers)
        return json.loads(query_result.content)

    @classmethod
    def get_milestones(cls, base_url: str, user: str, password: str, project_id: int) -> dict:
        """Retrieve all available milestones under the given project.

        :param base_url: TestRail's base url
        :param user: user's session id
        :param password: user's password
        :param project_id: target project id

        :return: all available milestones information
        """
        url = cls._construct_url(base_url, cls.API_VERSION, uri=f"get_milestones/{project_id}")
        headers = cls._construct_header(user, password, ContentType.JSON.value)
        query_result = HttpRequest.get(url=url, headers=headers)
        return json.loads(query_result.content)

    @classmethod
    def add_result_for_case(
        cls,
        base_url: str,
        user: str,
        password: str,
        run_id: int,
        case_id: id,
        data: dict,
    ) -> dict:
        """Add the given result to the given test case for the
        corresponding run.

        :param base_url: TestRail's base url
        :param user: user's session id
        :param password: user's password
        :param run_id: TestRail's run id
        :param case_id: TestRail's case id
        :param data: result to add (status, report...)

        :return: new test result
        """
        url = cls._construct_url(
            base_url,
            cls.API_VERSION,
            uri=f"add_result_for_case/{run_id}/{case_id}",
        )
        headers = cls._construct_header(user, password, ContentType.JSON.value)
        query_result = HttpRequest.post(url=url, headers=headers, data=data)
        return json.loads(query_result.content)

    @classmethod
    def add_run(
        cls,
        base_url: str,
        user: str,
        password: str,
        project_id: int,
        data: dict,
    ) -> dict:
        """Create a brand new TestRail run.

        :param base_url: TestRail's base url
        :param user: user's session id
        :param password: user's password
        :param project_id: TestRail's project id
        :param data: additional data to add to the run (milestonem,
            case ids,...)

        :return: the new test run
        """
        url = cls._construct_url(base_url, cls.API_VERSION, uri=f"add_run/{project_id}")
        headers = cls._construct_header(user, password, ContentType.JSON.value)
        query_result = HttpRequest.post(url=url, headers=headers, data=data)
        return json.loads(query_result.content)



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
TRC file handler based on Python CAN
************************************

:module: trc_handler

:synopsis: Additional layer for python-can trc reader and write to handle can FD

.. currentmodule:: trc_handler

"""
import logging
from datetime import datetime, timedelta
from typing import List, Optional, Tuple

try:
    from can import Message
    from can.io.trc import TRCFileVersion, TRCReader, TRCWriter
    from can.util import channel2int, dlc2len, len2dlc
except ImportError as e:
    raise ImportError(f"{e.name} dependency missing, consider installing pykiso with 'pip install pykiso[can]'")


log = logging.getLogger(__name__)


NO_ARBITRATION_ID_TYPES = {"EC", "ER", "ST"}
TRC_V2_FRAME_TYPES = {"DT", "FD", "FB", "FE", "BI", "RR", "EC", "ER", "ST"}


class TypedMessage(Message):
    """Message with type attribute added and that can handle empty arbitration id.
    Necessary because python-can does not handle types from CAN FD and information
    contained in Message are not enough to deduce it.
    Some of those types (ex: ST) also have no arbitration id which cannot be handle by
    python-can Message object.
    """

    __slots__ = "type"

    def __init__(self, type: str, **kwargs):
        """Create a CAN message with message type information.

        :param type: type of CAN message (ex:DT)
        :param kwargs: all of the keyword arguments expected by can.Message
        """
        super().__init__(**kwargs)
        self.type = type

    def __repr__(self) -> str:
        """String representation that can handle messages with no arbitration id

        :return: string representation of the message.
        """
        if self.type in NO_ARBITRATION_ID_TYPES:
            args = [
                f"timestamp={self.timestamp}",
                f"arbitration_id={self.arbitration_id}",
                f"is_extended_id={self.is_extended_id}",
            ]
        else:
            args = [
                f"timestamp={self.timestamp}",
                f"arbitration_id={self.arbitration_id:#x}",
                f"is_extended_id={self.is_extended_id}",
            ]

        if not self.is_rx:
            args.append("is_rx=False")

        if self.is_remote_frame:
            args.append(f"is_remote_frame={self.is_remote_frame}")

        if self.is_error_frame:
            args.append(f"is_error_frame={self.is_error_frame}")

        if self.channel is not None:
            args.append(f"channel={self.channel!r}")

        data = [f"{byte:#02x}" for byte in self.data]
        args += [f"dlc={self.dlc}", f"data=[{', '.join(data)}]"]

        if self.is_fd:
            args.append("is_fd=True")
            args.append(f"bitrate_switch={self.bitrate_switch}")
            args.append(f"error_state_indicator={self.error_state_indicator}")

        return f"can.Message({', '.join(args)})"


class TRCReaderCanFD(TRCReader):
    """Parse trc file to extract messages information
    Unlike the base TRCReader class, also retrieve message type
    """

    def _nomalize_cols(self, cols: Tuple[str] | List[str]) -> Tuple[str] | List[str]:
        """Insert missing columns for specific messages (Hardware Status, Error Frames
        and Error Counter Changes) to avoid

        :param cols: message information with missing columns

        :return: normalized message
        """
        tuple_used = isinstance(cols, tuple)
        cols = list(cols) if tuple_used else cols
        type_ = cols[self.columns["T"]]
        if type_ in NO_ARBITRATION_ID_TYPES:
            if self.file_version == TRCFileVersion.V2_0:
                cols.insert(self.columns["I"], "    ")
            else:
                cols.insert(self.columns["I"], "   -")
            if "l" in self.columns:
                cols.insert(self.columns["l"], " ")
            if "L" in self.columns:
                cols.insert(self.columns["L"], " ")
        return tuple(cols) if tuple_used else cols

    def _parse_cols_v2_x(self, cols: List[str] | Tuple[str, ...]) -> Optional[TypedMessage]:
        """Parse columns for file version 2.0 and 2.1.

        :param cols: list of columns to parse

        :return: typed message or None if message type is not supported
        """
        dtype = cols[self.columns["T"]]
        cols = self._nomalize_cols(cols)

        if dtype in TRC_V2_FRAME_TYPES:
            return self._parse_msg_v2_x(cols)
        else:
            log.info("TRCReader: Unsupported type '%s'", dtype)
            return None

    def _parse_msg_v2_x(self, cols: List[str] | Tuple[str]) -> Optional[Message]:
        """Parse messages for file version 2.0 and 2.1.

        :param cols: list of columns to parse

        return: typed message
        """
        type_ = cols[self.columns["T"]]
        bus = self.columns.get("B", None)

        # For 2.0 ST / ER and EC messages dlc and length are not in the trace but length is fixed
        if type_ in NO_ARBITRATION_ID_TYPES and self.file_version == TRCFileVersion.V2_0:
            if type_ == "ST":
                length = 4
            elif type_ == "ER":
                length = 5
            else:
                length = 2
            dlc = "  "
        else:
            if "l" in self.columns:
                length = int(cols[self.columns["l"]])
                dlc = len2dlc(length)
            elif "L" in self.columns:
                dlc = int(cols[self.columns["L"]])
                length = dlc2len(dlc)
            else:
                raise ValueError("No length/dlc columns present.")

        if isinstance(self.start_time, datetime):
            timestamp = (self.start_time + timedelta(milliseconds=float(cols[self.columns["O"]]))).timestamp()
        else:
            timestamp = float(cols[1]) / 1000

        if type_ not in NO_ARBITRATION_ID_TYPES or self.file_version == TRCFileVersion.V2_1:
            arbitration_id = int(cols[self.columns["I"]], 16)
            is_extended_id = len(cols[self.columns["I"]]) > 4
        # No arbitration id for ST / ER and EC messages
        else:
            arbitration_id = cols[self.columns["I"]]
            is_extended_id = False

        channel = int(cols[bus]) if bus is not None else 1
        dlc = dlc
        data = bytearray([int(cols[i + self.columns["D"]], 16) for i in range(length)])
        is_rx = cols[self.columns["d"]] == "Rx"
        is_fd = type_ in ["FD", "FB", "FE", "BI"]
        bitrate_switch = type_ in ["FB", " FE"]
        error_state_indicator = type_ in ["FE", "BI"]
        is_error_frame = type_ == "ER"
        is_remote_frame = type_ == "RR"

        return TypedMessage(
            type_,
            timestamp=timestamp,
            arbitration_id=arbitration_id,
            is_extended_id=is_extended_id,
            is_error_frame=is_error_frame,
            is_remote_frame=is_remote_frame,
            channel=channel,
            dlc=dlc,
            data=data,
            is_fd=is_fd,
            is_rx=is_rx,
            bitrate_switch=bitrate_switch,
            error_state_indicator=error_state_indicator,
        )


class TRCWriterCanFD(TRCWriter):
    """Logs CAN FD data to text file (.trc)"""

    # Type has been added to FORMAT_MESSAGE >= 2.0
    FORMAT_MESSAGE_V2_1 = "{msgnr:>7} {time:13.3f} {type:>2} {channel:>2} {id:>8} {dir:>2} - {dlc:<4} {data}"
    FORMAT_MESSAGE_V2_0 = "{msgnr:>7} {time:13.3f} {type:>2} {id:>8} {dir:>2} {dlc:<2} {data}"
    FILE_VERSION_TO_FORMAT = {
        TRCFileVersion.V1_0: TRCWriter.FORMAT_MESSAGE_V1_0,
        TRCFileVersion.V2_0: FORMAT_MESSAGE_V2_0,
        TRCFileVersion.V2_1: FORMAT_MESSAGE_V2_1,
    }

    def _format_message_init(self, msg: TypedMessage, channel: int) -> str:
        """Pick message format from file version and format initial message

        :return: message information formatted in a string
        """
        # Fix error from python can -> message number should start at one not zero
        self.msgnr = 1

        try:
            self._msg_fmt_string = self.FILE_VERSION_TO_FORMAT[self.file_version]
        except KeyError as e:
            raise NotImplementedError(f"File format {e} is not supported")

        self._format_message = self._format_message_by_format
        return self._format_message_by_format(msg, channel)

    def _format_message_by_format(self, msg: TypedMessage, channel: int):
        """Format messages

        :return: message information formatted in a string
        """
        if msg.type in NO_ARBITRATION_ID_TYPES:
            arb_id = f"{msg.arbitration_id}"
        else:
            if msg.is_extended_id:
                arb_id = f"{msg.arbitration_id:07X}"
            else:
                arb_id = f"{msg.arbitration_id:04X}"

        data = [f"{byte:02X}" for byte in msg.data]

        # For the time python-can was doing subtraction with the first message timestamp
        # which is incorrect. It should be with the start time of the trace otherwise the
        # first message will always have an offset of zero.
        if self.file_version == TRCFileVersion.V1_0:
            serialized = self._msg_fmt_string.format(
                msgnr=self.msgnr,
                time=(msg.timestamp - self.first_timestamp) * 1000,
                channel=channel,
                id=arb_id,
                dir="Rx" if msg.is_rx else "Tx",
                dlc=msg.dlc,
                data=" ".join(data),
            )
        else:
            serialized = self._msg_fmt_string.format(
                msgnr=self.msgnr,
                time=(msg.timestamp - self.first_timestamp) * 1000,
                type=msg.type,
                channel=channel,
                id=arb_id,
                dir="Rx" if msg.is_rx else "Tx",
                dlc=msg.dlc,
                data=" ".join(data),
            )
        return serialized

    def set_header_data(self, header_data: List[str]):
        """Get all header info so trc writter can write it when needed

        :param header_data: list of all header lines
        """
        self.header_data = header_data

    def write_header(self, timestamp: float = 0) -> None:
        """Modified write header that just write previously given header data

        :param timestamp: unused parameter kept to avoid breaking changes
        """
        if self.file_version in self.FILE_VERSION_TO_FORMAT.keys():
            self.file.writelines(line for line in self.header_data)
        else:
            raise NotImplementedError("File format is not supported")
        self.header_written = True

    def on_message_received(self, msg: TypedMessage, trace_start_time: float) -> None:
        """Log a message in the trace. Handles CAN FD.

        In python can the reference is the timestamp of the first message (the first
        message offset is always zero), to fix this we take the beginning of the trace.

        :param msg: typed message to log
        :param trace_start_time: timestamp of the start of the trace
        """

        if self.first_timestamp is None:
            self.first_timestamp = trace_start_time

        channel = channel2int(msg.channel)
        if channel is None:
            channel = self.channel
        else:
            # Many interfaces start channel numbering at 0 which is invalid
            channel += 1
        serialized = self._format_message(msg, channel)
        self.msgnr += 1
        self.log_event(serialized, msg.timestamp)



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
acroname auxiliary plugin
*************************

:module: acroname_auxiliary

:synopsis: implementation of existing AcronameAuxiliary for
    Robot framework usage.

.. currentmodule:: acroname_auxiliary

"""

from typing import Optional

from robot.api.deco import keyword, library

from ..auxiliaries.acroname_auxiliary import AcronameAuxiliary as AcroAux
from .aux_interface import RobotAuxInterface


@library(version="0.1.0")
class AcronameAuxiliary(RobotAuxInterface):
    """Robot framework plugin for AcronameAuxiliary."""

    ROBOT_LIBRARY_SCOPE = "SUITE"

    def __init__(self):
        """Initialize attributes."""
        super().__init__(aux_type=AcroAux)

    @keyword(name="Set port enable")
    def set_port_enable(self, aux_alias: str, port: int) -> int:
        """Enable power and data lines for a USB port.

        :param aux_alias: auxiliary's alias
        :param port: the USB port number
        :return: brainstem error code. 0 if no error.
        """
        aux = self._get_aux(aux_alias)
        return aux.set_port_enable(port)

    @keyword(name="Set port disable")
    def set_port_disable(self, aux_alias: str, port: int) -> int:
        """Disable power and data lines for a USB port.

        :param aux_alias: auxiliary's alias
        :param port: the USB port number
        :return: brainstem error code. 0 if no error.
        """
        aux = self._get_aux(aux_alias)
        return aux.set_port_disable(port)

    @keyword(name="Get port current")
    def get_port_current(self, aux_alias: str, port: int, unit: str = "A") -> Optional[float]:
        """Get the current through the power line for selected usb port.

        :param aux_alias: auxiliary's alias
        :param port: the USB port number
        :param unit: unit of the result in "uA", "mA" or "A". Default "A"
        :return: port current for given unit. None if unit is not supported.
        """
        aux = self._get_aux(aux_alias)
        return aux.get_port_current(port, unit)

    @keyword(name="Get port voltage")
    def get_port_voltage(self, aux_alias: str, port: int, unit: str = "V") -> Optional[float]:
        """Get the voltage of the selected usb port.

        :param aux_alias: auxiliary's alias
        :param port: the USB port number
        :param unit: unit of the result in "uV", "mV" or "V". Default "V"
        :return: port voltage for given unit. None if unit is not supported.
        """
        aux = self._get_aux(aux_alias)
        return aux.get_port_voltage(port, unit)

    @keyword(name="Get port current limit")
    def get_port_current_limit(self, aux_alias: str, port: int, unit: str = "A") -> Optional[float]:
        """Get the current limit for the port.

        :param aux_alias: auxiliary's alias
        :param port: the USB port number
        :param unit: unit of the result in "uA", "mA" or "A". Default "A"
        :return: port current limit for given unit. None if unit is not supported.
        """
        aux = self._get_aux(aux_alias)
        return aux.get_port_current_limit(port, unit)

    @keyword(name="Set port current limit")
    def set_port_current_limit(self, aux_alias: str, port: int, amps: float, unit: str = "A") -> None:
        """Set the current limit for the port. If the set limit is not achievable,
        devices will round down to the nearest available current limit setting.

        :param aux_alias: auxiliary's alias
        :param port: the USB port number
        :param amps: value for port current to set in "uA", "mA" or "A". Default "A"
        :param unit: unit for the value to set. Default Ampere
        """
        aux = self._get_aux(aux_alias)
        return aux.set_port_current_limit(port, amps, unit)



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
pykiso Control Message Protocol
*******************************

:module: can_message

:synopsis: Message transfer by can

.. currentmodule:: message
"""

from typing import Any


class CanMessage:
    def __init__(self, name: str, signals: dict[str, Any], timestamp: float) -> None:
        self.name = name
        self.signals = signals
        self.timestamp = timestamp



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
Communication Channel via socket
********************************

:module: cc_socket

:synopsis: connector for communication via socket

.. currentmodule:: cc_socket

"""
import logging
import socket
from typing import Dict, Optional, Union

from pykiso import CChannel

log = logging.getLogger(__name__)


class CCTcpip(CChannel):
    """Connector channel used to communicate via socket"""

    def __init__(self, dest_ip: str, dest_port: int, max_msg_size: int = 256, **kwargs):
        """Initialize channel settings.

        :param dest_ip: destination ip address
        :param dest_port: destination port
        :param max_msg_size: the maximum amount of data to be received
            at once
        """
        super().__init__(**kwargs)
        self.dest_ip = dest_ip
        self.dest_port = int(dest_port)
        self.socket: Optional[socket.socket] = None
        self.max_msg_size = max_msg_size
        # Set a timeout to send the signal to the GIL to change thread.
        # In case of a multi-threading system, all tasks will be called one after the other.
        self.timeout = 1e-6

    def _cc_open(self) -> None:
        """Connect to socket with configured port and IP address."""
        if self.socket:
            log.internal_info(
                "`CCTcpip._cc_open` called again; already connected "
                f"to socket at address {self.dest_ip} port {self.dest_port}"
            )
            return
        log.internal_info(f"Connection to socket at address {self.dest_ip} port {self.dest_port}")
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.socket.settimeout(3)
        self.socket.connect((self.dest_ip, self.dest_port))

    def _cc_close(self) -> None:
        """Close UDP socket."""
        if self.socket:
            log.internal_info(f"Disconnect from socket at address {self.dest_ip}, port {self.dest_port}")
            self.socket.close()
            self.socket = None

    def _cc_send(self, msg: bytes or str, **kwargs) -> None:
        """Send a message via socket.

        :param msg: message to send
        :param kwargs: not used
        """
        if not self.socket:
            raise RuntimeError("Channel must be opened before messages can be sent")
        if isinstance(msg, str):
            msg = msg.encode()
        log.internal_debug(f"Sending {msg} via socket to {self.dest_ip}")
        self.socket.send(msg)

    def _cc_receive(self, timeout=0.01) -> Dict[str, Optional[bytes]]:
        """Read message from socket.

        :param timeout: time in second to wait for reading a message

        :return: Message if successful, otherwise none
        """
        if not self.socket:
            raise RuntimeError("Channel must be opened before messages can be received")

        self.socket.settimeout(timeout or self.timeout)

        try:
            msg_received = self.socket.recv(self.max_msg_size)
            log.internal_debug(f"Socket at {self.dest_ip} received: {msg_received}")
        except socket.timeout:
            log.exception(f"encountered timeout error while receiving message via {self}")
            return {"msg": None}
        except Exception:
            log.exception(f"encountered error while receiving message via {self}")
            return {"msg": None}

        return {"msg": msg_received}



##########################################################################
# Copyright (c) 2010-2022 Robert Bosch GmbH
# This program and the accompanying materials are made available under the
# terms of the Eclipse Public License 2.0 which is available at
# http://www.eclipse.org/legal/epl-2.0.
#
# SPDX-License-Identifier: EPL-2.0
##########################################################################

"""
Device Under Test Auxiliary
***************************

:module: DUTAuxiliary

:synopsis: The Device Under Test auxiliary allow to flash and run test
    on the target using the connector provided.

.. currentmodule:: dut_auxiliary

"""
from __future__ import annotations

import functools
import logging
import queue
from typing import Callable, Optional

from pykiso import CChannel, Flasher, Message, message
from pykiso.auxiliary import AuxiliaryInterface, close_connector, flash_target, open_connector

log = logging.getLogger(__name__)

MESSAGE_TYPE = message.MessageType
COMMAND_TYPE = message.MessageCommandType
REPORT_TYPE = message.MessageReportType


def retry_command(tries: int) -> Callable:
    """Force to resend the command a define number of times in case of
    failure.

    :param tries: maximum number of try to get the acknowledgement from
        the DUT

    :return: inner decorator function
    """

    def inner_retry(func: Callable) -> Callable:
        """Inner decorator function.

        :param func: decorated method

        :return: decorator inner function
        """

        @functools.wraps(func)
        def handle_ack(*arg: tuple, **kwargs: dict) -> bool:
            """Based on the run_command method return, just resend the
            command.

            :param arg: positional arguments
            :param kwargs: named arguments

            :return: True if the command was acknowledged otherwise
                False
            """
            for idx in range(tries):
                log.internal_info(f"command try n: {idx}")
                if func(*arg, **kwargs):
                    return True
            return False

        return handle_ack

    return inner_retry


def check_acknowledgement(func: Callable) -> Callable:
    """Check ifthe DUT has acknowledge the previous sent command.

    :param func: decorated method

    :return: decorator inner function
    """

    @functools.wraps(func)
    def inner_check(self, *arg: tuple, **kwargs: dict) -> bool:
        """Check if an ACK message was received and if the token is
        valid.

        :param self: aux instance
        :param arg: positional arguments
        :param kwargs: named arguments

        :return: True if the command was acknowledged otherwise
            False
        """
        response = func(self, *arg, **kwargs)

        # nothing was received
        if response is None:
            log.error("Nothing received from the DUT for the current request!")
            return False

        is_ack = self.current_cmd.check_if_ack_message_is_matching(response)

        # invalid token or received message not type of ACK
        if is_ack is False:
            log.internal_warning(f"Received {response} not matching {self.current_cmd}!")
            return False

        log.internal_info("Command was acknowledged by the DUT!")
        return True

    return inner_check


def restart_aux(func: Callable) -> Callable:
    """Force he auxiliary restart if the command is not acknowledge

    :param func: decorated method

    :return: decorator inner function
    """

    @functools.wraps(func)
    def inner_start(self, *arg: tuple, **kwargs: dict) -> bool:
        """Based on the run_command method return, force the auxiliary
        to create a brand new communication stream with the DUT (call of
        delete/create instance).

        :param self: aux instance
        :param arg: positional arguments
        :param kwargs: named arguments

        :return: True if the command was acknowledged otherwise
            False
        """
        state = func(self, *arg, **kwargs)

        # restart the whole auxiliary to have a brand new connection
        # with the DUT
        if state is False:
            self.delete_instance()
            self.create_instance()
        return state

    return inner_start


class DUTAuxiliary(AuxiliaryInterface):
    """Device Under Test(DUT) auxiliary implementation."""

    def __init__(
        self,
        com: CChannel = None,
        flash: Flasher = None,
        **kwargs,
    ):
        """Constructor.

        :param name: Alias of the auxiliary instance
        :param com: Communication connector
        :param flash: flash connector
        """
        super().__init__(is_proxy_capable=False, tx_task_on=True, rx_task_on=True, **kwargs)
        self.channel = com
        self.flash = flash
        self.current_cmd = None

    @retry_command(tries=1)
    @check_acknowledgement
    def send_ping_command(self, timeout: int) -> bool:
        """Send ping command to the DUT.

        :param timeout: how long to wait for an acknowledgement from the
            DUT (seconds)

        :return: True if the command is acknowledged otherwise False
        """

        # TODO: remove it when DUT start sending message at start-up
        # clean the reception queue due to a bug on embedded TestApp,
        while not self.queue_out.empty():
            self.queue_out.get_nowait()

        self.current_cmd = message.Message(MESSAGE_TYPE.COMMAND, COMMAND_TYPE.PING)
        log.internal_info(f"send ping request: {self.current_cmd}")
        return self.run_command(
            cmd_message=self.current_cmd,
            cmd_data=None,
            blocking=True,
            timeout_in_s=timeout,
        )

    @retry_command(tries=2)
    @check_acknowledgement
    def send_fixture_command(self, command: message.Message, timeout: int) -> bool:
        """Send command related to test execution (test case setup,
        test case run...) to the DUT.

        :param command: command to execute on DUT side
        :param timeout: how long to wait for an acknowledgement from the
            DUT (seconds)

        :return: True if the command is acknowledged otherwise False
        """
        self.current_cmd = command
        log.internal_info(f"send request: {self.current_cmd}")
        return self.run_command(
            cmd_message=self.current_cmd,
            cmd_data=None,
            blocking=True,
            timeout_in_s=timeout,
        )

    @retry_command(tries=2)
    @restart_aux
    @check_acknowledgement
    def send_abort_command(self, timeout: int):
        """Send abort command to the DUT.

        .. warning:: if the DUT doesn't acknowledge to the abort command
            the auxiliary is restarted, a brand new connection is
            established

        :param timeout: how long to wait for an acknowledgement from the
            DUT (seconds)

        :return: True if the command is acknowledged otherwise False
        """
        self.current_cmd = message.Message(MESSAGE_TYPE.COMMAND, COMMAND_TYPE.ABORT)
        log.internal_info(f"send abort request: {self.current_cmd}")
        return self.run_command(
            cmd_message=self.current_cmd,
            cmd_data=None,
            blocking=True,
            timeout_in_s=timeout,
        )

    def create_instance(self) -> bool:
        """Create DUT auxiliary instance.

        Overridden from base interface in order to use the TX and RX
        tasks, and not duplicate auxiliary method.
        Execute directly the ping-pong to initiate the communication
        with the DUT.

        :return: True if the auxiliary is created and ping-pong
            successful otherwise False

        :raises AuxiliaryCreationError: if instance creation failed
        """
        super().create_instance()
        return self.send_ping_command(timeout=2)

    @flash_target
    @open_connector
    def _create_auxiliary_instance(self) -> bool:
        """Create auxiliary instance flash and connect.

        :return: True if everything was successful otherwise False
        """
        log.internal_info("Auxiliary instance created")
        return True

    @close_connector
    def _delete_auxiliary_instance(self) -> bool:
        """Close the connector communication.

        :return: always True
        """
        log.internal_info("Auxiliary instance deleted")
        return True

    def evaluate_response(self, response: message.Message) -> bool:
        """Evaluate if the received message is knowned and type of
        report.

        .. note:: if a log message type is received just log it

        :param response: reeceived response

        :return: True if the response is a report otherwise False
        """
        if response.get_message_type() == MESSAGE_TYPE.REPORT:
            self.evaluate_report(report_msg=response)
            return True

        if response.get_message_type() == MESSAGE_TYPE.LOG:
            log.internal_info(f"Logging message received from {self.name}: {response}")
            return False

        if response.get_message_type() == MESSAGE_TYPE.ACK:
            log.internal_warning(f"ACK message received from {self.name}: {response}")
            return False

        log.internal_warning(f"Message type unknown received from {self.name}: {response}")
        return False

    def evaluate_report(self, report_msg: message.Message) -> None:
        """Evaluate the report type and log the appropriated message.

        :param report_msg: reeceived report message
        """
        if report_msg.get_message_sub_type() == REPORT_TYPE.TEST_FAILED:
            log.error(f"Report with verdict FAILED received from {self.name} : {report_msg}")
        elif report_msg.get_message_sub_type() == REPORT_TYPE.TEST_PASS:
            log.internal_info(f"Report with verdict PASS received from {self.name} : {report_msg}")

        elif report_msg.get_message_sub_type() == REPORT_TYPE.TEST_NOT_IMPLEMENTED:
            log.internal_warning(f"Report with verdict NOT IMPLEMENTED from {self.name} : {report_msg}")
        else:
            log.error(f"Report type unknown received from {self.name} : {report_msg}")

    def wait_and_get_report(self, blocking: bool = False, timeout_in_s: int = 0) -> Optional[Message]:
        """Wait for the report coming from the DUT.

        :param blocking: True: wait for timeout to expire, False: return
             immediately
        :param timeout_in_s: if blocking, wait the defined time in
            seconds

        :return: if a report is received return it otherwise None
        """
        try:
            is_report = False
            while not is_report:
                response = self.queue_out.get(blocking, timeout_in_s)
                is_report = self.evaluate_response(response)
            return response
        except queue.Empty:
            return None

    def _run_command(
        self,
        cmd_message: message.Message,
        cmd_data: bytes = None,
    ) -> None:
        """Simply send the given command using the associated channel.

        :param cmd_message: command to send
        :param cmd_data: not use
        """
        try:
            # Serialize and send the message
            self.channel.cc_send(msg=cmd_message.serialize())
        except Exception:
            log.exception(f"encountered error while sending message '{cmd_message}' to {self.channel}")

    def _receive_message(self, timeout_in_s: float) -> None:
        """Get message from the device under test.

        :param timeout_in_s: Time in seconds to wait for an answer
        """

        recv_response = self.channel.cc_receive(timeout_in_s)
        response = recv_response.get("msg")

        if response is None:
            return
        elif isinstance(response, str):
            response = response.encode()
        response = Message.parse_packet(response)

        # If a message was received just automatically acknowledge it
        # and populate the queue_out
        if response.msg_type != MESSAGE_TYPE.ACK:
            ack_cmd = response.generate_ack_message(message.MessageAckType.ACK)
            try:
                self.channel.cc_send(msg=ack_cmd.serialize())
            except Exception:
                log.exception(f"encountered error while sending acknowledge message for {response}!")
        self.queue_out.put(response)
